@@ -8,7 +8,7 @@
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/access/hash/hashpage.c,v 1.47.4.2 2007/04/19 20:24:28 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/access/hash/hashpage.c,v 1.52 2005/10/15 02:49:08 momjian Exp $
  *
  * NOTES
  *	  Postgres hash pages look like ordinary relation pages.  The opaque
@@ -30,13 +30,11 @@
 
 #include "access/genam.h"
 #include "access/hash.h"
+#include "miscadmin.h"
 #include "storage/lmgr.h"
-#include "storage/smgr.h"
 #include "utils/lsyscache.h"
 
 
-static bool _hash_alloc_buckets(Relation rel, BlockNumber firstblock,
-								uint32 nblocks);
 static void _hash_splitbucket(Relation rel, Buffer metabuf,
 				  Bucket obucket, Bucket nbucket,
 				  BlockNumber start_oblkno,
@@ -104,9 +102,10 @@ _hash_droplock(Relation rel, BlockNumber whichlock, int access)
  *		requested buffer and its reference count has been incremented
  *		(ie, the buffer is "locked and pinned").
  *
- *		P_NEW is disallowed because this routine should only be used
- *		to access pages that are known to be before the filesystem EOF.
- *		Extending the index should be done with _hash_getnewbuf.
+ *		XXX P_NEW is not used because, unlike the tree structures, we
+ *		need the bucket blocks to be at certain block numbers.	we must
+ *		depend on the caller to call _hash_pageinit on the block if it
+ *		knows that this is a new block.
  */
 Buffer
 _hash_getbuf(Relation rel, BlockNumber blkno, int access)
@@ -125,51 +124,6 @@ _hash_getbuf(Relation rel, BlockNumber blkno, int access)
 	return buf;
 }
 
-/*
- *	_hash_getnewbuf() -- Get a new page at the end of the index.
- *
- *		This has the same API as _hash_getbuf, except that we are adding
- *		a page to the index, and hence expect the page to be past the
- *		logical EOF.  (However, we have to support the case where it isn't,
- *		since a prior try might have crashed after extending the filesystem
- *		EOF but before updating the metapage to reflect the added page.)
- *
- *		It is caller's responsibility to ensure that only one process can
- *		extend the index at a time.
- *
- *		All call sites should call _hash_pageinit on the returned page.
- *		Also, it's difficult to imagine why access would not be HASH_WRITE.
- */
-Buffer
-_hash_getnewbuf(Relation rel, BlockNumber blkno, int access)
-{
-	BlockNumber	nblocks = RelationGetNumberOfBlocks(rel);
-	Buffer		buf;
-
-	if (blkno == P_NEW)
-		elog(ERROR, "hash AM does not use P_NEW");
-	if (blkno > nblocks)
-		elog(ERROR, "access to noncontiguous page in hash index \"%s\"",
-			 RelationGetRelationName(rel));
-
-	/* smgr insists we use P_NEW to extend the relation */
-	if (blkno == nblocks)
-	{
-		buf = ReadBuffer(rel, P_NEW);
-		if (BufferGetBlockNumber(buf) != blkno)
-			elog(ERROR, "unexpected hash relation size: %u, should be %u",
-				 BufferGetBlockNumber(buf), blkno);
-	}
-	else
-		buf = ReadBuffer(rel, blkno);
-
-	if (access != HASH_NOLOCK)
-		LockBuffer(buf, access);
-
-	/* ref count and lock type are correct */
-	return buf;
-}
-
 /*
  *	_hash_relbuf() -- release a locked buffer.
  *
@@ -286,13 +240,13 @@ _hash_metapinit(Relation rel)
 			 RelationGetRelationName(rel));
 
 	/*
-	 * Determine the target fill factor (tuples per bucket) for this
-	 * index. The idea is to make the fill factor correspond to pages
-	 * about 3/4ths full.  We can compute it exactly if the index datatype
-	 * is fixed-width, but for var-width there's some guessing involved.
+	 * Determine the target fill factor (tuples per bucket) for this index.
+	 * The idea is to make the fill factor correspond to pages about 3/4ths
+	 * full.  We can compute it exactly if the index datatype is fixed-width,
+	 * but for var-width there's some guessing involved.
 	 */
 	data_width = get_typavgwidth(RelationGetDescr(rel)->attrs[0]->atttypid,
-							 RelationGetDescr(rel)->attrs[0]->atttypmod);
+								 RelationGetDescr(rel)->attrs[0]->atttypmod);
 	item_width = MAXALIGN(sizeof(HashItemData)) + MAXALIGN(data_width) +
 		sizeof(ItemIdData);		/* include the line pointer */
 	ffactor = (BLCKSZ * 3 / 4) / item_width;
@@ -300,13 +254,7 @@ _hash_metapinit(Relation rel)
 	if (ffactor < 10)
 		ffactor = 10;
 
-	/*
-	 * We initialize the metapage, the first two bucket pages, and the
-	 * first bitmap page in sequence, using _hash_getnewbuf to cause
-	 * smgrextend() calls to occur.  This ensures that the smgr level
-	 * has the right idea of the physical index length.
-	 */
-	metabuf = _hash_getnewbuf(rel, HASH_METAPAGE, HASH_WRITE);
+	metabuf = _hash_getbuf(rel, HASH_METAPAGE, HASH_WRITE);
 	pg = BufferGetPage(metabuf);
 	_hash_pageinit(pg, BufferGetPageSize(metabuf));
 
@@ -341,15 +289,14 @@ _hash_metapinit(Relation rel)
 	metap->hashm_procid = index_getprocid(rel, 1, HASHPROC);
 
 	/*
-	 * We initialize the index with two buckets, 0 and 1, occupying
-	 * physical blocks 1 and 2.  The first freespace bitmap page is in
-	 * block 3.
+	 * We initialize the index with two buckets, 0 and 1, occupying physical
+	 * blocks 1 and 2.	The first freespace bitmap page is in block 3.
 	 */
 	metap->hashm_maxbucket = metap->hashm_lowmask = 1;	/* nbuckets - 1 */
 	metap->hashm_highmask = 3;	/* (nbuckets << 1) - 1 */
 
-	MemSet((char *) metap->hashm_spares, 0, sizeof(metap->hashm_spares));
-	MemSet((char *) metap->hashm_mapp, 0, sizeof(metap->hashm_mapp));
+	MemSet(metap->hashm_spares, 0, sizeof(metap->hashm_spares));
+	MemSet(metap->hashm_mapp, 0, sizeof(metap->hashm_mapp));
 
 	metap->hashm_spares[1] = 1; /* the first bitmap page is only spare */
 	metap->hashm_ovflpoint = 1;
@@ -360,7 +307,7 @@ _hash_metapinit(Relation rel)
 	 */
 	for (i = 0; i <= 1; i++)
 	{
-		buf = _hash_getnewbuf(rel, BUCKET_TO_BLKNO(metap, i), HASH_WRITE);
+		buf = _hash_getbuf(rel, BUCKET_TO_BLKNO(metap, i), HASH_WRITE);
 		pg = BufferGetPage(buf);
 		_hash_pageinit(pg, BufferGetPageSize(buf));
 		pageopaque = (HashPageOpaque) PageGetSpecialPointer(pg);
@@ -373,7 +320,8 @@ _hash_metapinit(Relation rel)
 	}
 
 	/*
-	 * Initialize first bitmap page
+	 * Initialize first bitmap page.  Can't do this until we create the first
+	 * two buckets, else smgr will complain.
 	 */
 	_hash_initbitmap(rel, metap, 3);
 
@@ -418,15 +366,14 @@ _hash_expandtable(Relation rel, Buffer metabuf)
 	 * Obtain the page-zero lock to assert the right to begin a split (see
 	 * README).
 	 *
-	 * Note: deadlock should be impossible here. Our own backend could only
-	 * be holding bucket sharelocks due to stopped indexscans; those will
-	 * not block other holders of the page-zero lock, who are only
-	 * interested in acquiring bucket sharelocks themselves.  Exclusive
-	 * bucket locks are only taken here and in hashbulkdelete, and neither
-	 * of these operations needs any additional locks to complete.	(If,
-	 * due to some flaw in this reasoning, we manage to deadlock anyway,
-	 * it's okay to error out; the index will be left in a consistent
-	 * state.)
+	 * Note: deadlock should be impossible here. Our own backend could only be
+	 * holding bucket sharelocks due to stopped indexscans; those will not
+	 * block other holders of the page-zero lock, who are only interested in
+	 * acquiring bucket sharelocks themselves.	Exclusive bucket locks are
+	 * only taken here and in hashbulkdelete, and neither of these operations
+	 * needs any additional locks to complete.	(If, due to some flaw in this
+	 * reasoning, we manage to deadlock anyway, it's okay to error out; the
+	 * index will be left in a consistent state.)
 	 */
 	_hash_getlock(rel, 0, HASH_EXCLUSIVE);
 
@@ -437,38 +384,29 @@ _hash_expandtable(Relation rel, Buffer metabuf)
 	_hash_checkpage(rel, (Page) metap, LH_META_PAGE);
 
 	/*
-	 * Check to see if split is still needed; someone else might have
-	 * already done one while we waited for the lock.
+	 * Check to see if split is still needed; someone else might have already
+	 * done one while we waited for the lock.
 	 *
-	 * Make sure this stays in sync with_hash_doinsert()
+	 * Make sure this stays in sync with _hash_doinsert()
 	 */
 	if (metap->hashm_ntuples <=
 		(double) metap->hashm_ffactor * (metap->hashm_maxbucket + 1))
 		goto fail;
 
-	/*
-	 * Can't split anymore if maxbucket has reached its maximum possible value.
-	 *
-	 * Ideally we'd allow bucket numbers up to UINT_MAX-1 (no higher because
-	 * the calculation maxbucket+1 mustn't overflow).  Currently we restrict
-	 * to half that because of overflow looping in _hash_log2() and
-	 * insufficient space in hashm_spares[].  It's moot anyway because an
-	 * index with 2^32 buckets would certainly overflow BlockNumber and
-	 * hence _hash_alloc_buckets() would fail, but if we supported buckets
-	 * smaller than a disk block then this would be an independent constraint.
-	 */
-	if (metap->hashm_maxbucket >= (uint32) 0x7FFFFFFE)
-		goto fail;
-
 	/*
 	 * Determine which bucket is to be split, and attempt to lock the old
 	 * bucket.	If we can't get the lock, give up.
 	 *
 	 * The lock protects us against other backends, but not against our own
 	 * backend.  Must check for active scans separately.
+	 *
+	 * Ideally we would lock the new bucket too before proceeding, but if we are
+	 * about to cross a splitpoint then the BUCKET_TO_BLKNO mapping isn't
+	 * correct yet.  For simplicity we update the metapage first and then
+	 * lock.  This should be okay because no one else should be trying to lock
+	 * the new bucket yet...
 	 */
 	new_bucket = metap->hashm_maxbucket + 1;
-
 	old_bucket = (new_bucket & metap->hashm_lowmask);
 
 	start_oblkno = BUCKET_TO_BLKNO(metap, old_bucket);
@@ -480,48 +418,16 @@ _hash_expandtable(Relation rel, Buffer metabuf)
 		goto fail;
 
 	/*
-	 * Likewise lock the new bucket (should never fail).
+	 * Okay to proceed with split.	Update the metapage bucket mapping info.
 	 *
-	 * Note: it is safe to compute the new bucket's blkno here, even though
-	 * we may still need to update the BUCKET_TO_BLKNO mapping.  This is
-	 * because the current value of hashm_spares[hashm_ovflpoint] correctly
-	 * shows where we are going to put a new splitpoint's worth of buckets.
+	 * Since we are scribbling on the metapage data right in the shared buffer,
+	 * any failure in this next little bit leaves us with a big problem: the
+	 * metapage is effectively corrupt but could get written back to disk.	We
+	 * don't really expect any failure, but just to be sure, establish a
+	 * critical section.
 	 */
-	start_nblkno = BUCKET_TO_BLKNO(metap, new_bucket);
+	START_CRIT_SECTION();
 
-	if (_hash_has_active_scan(rel, new_bucket))
-		elog(ERROR, "scan in progress on supposedly new bucket");
-
-	if (!_hash_try_getlock(rel, start_nblkno, HASH_EXCLUSIVE))
-		elog(ERROR, "could not get lock on supposedly new bucket");
-
-	/*
-	 * If the split point is increasing (hashm_maxbucket's log base 2
-	 * increases), we need to allocate a new batch of bucket pages.
-	 */
-	spare_ndx = _hash_log2(new_bucket + 1);
-	if (spare_ndx > metap->hashm_ovflpoint)
-	{
-		Assert(spare_ndx == metap->hashm_ovflpoint + 1);
-		/*
-		 * The number of buckets in the new splitpoint is equal to the
-		 * total number already in existence, i.e. new_bucket.  Currently
-		 * this maps one-to-one to blocks required, but someday we may need
-		 * a more complicated calculation here.
-		 */
-		if (!_hash_alloc_buckets(rel, start_nblkno, new_bucket))
-		{
-			/* can't split due to BlockNumber overflow */
-			_hash_droplock(rel, start_oblkno, HASH_EXCLUSIVE);
-			_hash_droplock(rel, start_nblkno, HASH_EXCLUSIVE);
-			goto fail;
-		}
-	}
-
-	/*
-	 * Okay to proceed with split.	Update the metapage bucket mapping
-	 * info.
-	 */
 	metap->hashm_maxbucket = new_bucket;
 
 	if (new_bucket > metap->hashm_highmask)
@@ -534,22 +440,37 @@ _hash_expandtable(Relation rel, Buffer metabuf)
 	/*
 	 * If the split point is increasing (hashm_maxbucket's log base 2
 	 * increases), we need to adjust the hashm_spares[] array and
-	 * hashm_ovflpoint so that future overflow pages will be created
-	 * beyond this new batch of bucket pages.
+	 * hashm_ovflpoint so that future overflow pages will be created beyond
+	 * this new batch of bucket pages.
+	 *
+	 * XXX should initialize new bucket pages to prevent out-of-order page
+	 * creation?  Don't wanna do it right here though.
 	 */
+	spare_ndx = _hash_log2(metap->hashm_maxbucket + 1);
 	if (spare_ndx > metap->hashm_ovflpoint)
 	{
+		Assert(spare_ndx == metap->hashm_ovflpoint + 1);
 		metap->hashm_spares[spare_ndx] = metap->hashm_spares[metap->hashm_ovflpoint];
 		metap->hashm_ovflpoint = spare_ndx;
 	}
 
+	/* now we can compute the new bucket's primary block number */
+	start_nblkno = BUCKET_TO_BLKNO(metap, new_bucket);
+
+	Assert(!_hash_has_active_scan(rel, new_bucket));
+
+	if (!_hash_try_getlock(rel, start_nblkno, HASH_EXCLUSIVE))
+		elog(PANIC, "could not get lock on supposedly new bucket");
+
+	/* Done mucking with metapage */
+	END_CRIT_SECTION();
+
 	/*
 	 * Copy bucket mapping info now; this saves re-accessing the meta page
 	 * inside _hash_splitbucket's inner loop.  Note that once we drop the
-	 * split lock, other splits could begin, so these values might be out
-	 * of date before _hash_splitbucket finishes.  That's okay, since all
-	 * it needs is to tell which of these two buckets to map hashkeys
-	 * into.
+	 * split lock, other splits could begin, so these values might be out of
+	 * date before _hash_splitbucket finishes.	That's okay, since all it
+	 * needs is to tell which of these two buckets to map hashkeys into.
 	 */
 	maxbucket = metap->hashm_maxbucket;
 	highmask = metap->hashm_highmask;
@@ -583,72 +504,6 @@ fail:
 }
 
 
-/*
- * _hash_alloc_buckets -- allocate a new splitpoint's worth of bucket pages
- *
- * This does not need to initialize the new bucket pages; we'll do that as
- * each one is used by _hash_expandtable().  But we have to extend the logical
- * EOF to the end of the splitpoint; otherwise the first overflow page
- * allocated beyond the splitpoint will represent a noncontiguous access,
- * which can confuse md.c (and will probably be forbidden by future changes
- * to md.c).
- *
- * We do this by writing a page of zeroes at the end of the splitpoint range.
- * We expect that the filesystem will ensure that the intervening pages read
- * as zeroes too.  On many filesystems this "hole" will not be allocated
- * immediately, which means that the index file may end up more fragmented
- * than if we forced it all to be allocated now; but since we don't scan
- * hash indexes sequentially anyway, that probably doesn't matter.
- *
- * XXX It's annoying that this code is executed with the metapage lock held.
- * We need to interlock against _hash_getovflpage() adding a new overflow page
- * concurrently, but it'd likely be better to use LockRelationForExtension
- * for the purpose.  OTOH, adding a splitpoint is a very infrequent operation,
- * so it may not be worth worrying about.
- *
- * Returns TRUE if successful, or FALSE if allocation failed due to
- * BlockNumber overflow.
- */
-static bool
-_hash_alloc_buckets(Relation rel, BlockNumber firstblock, uint32 nblocks)
-{
-	BlockNumber	lastblock;
-	BlockNumber	endblock;
-	char		zerobuf[BLCKSZ];
-
-	lastblock = firstblock + nblocks - 1;
-
-	/*
-	 * Check for overflow in block number calculation; if so, we cannot
-	 * extend the index anymore.
-	 */
-	if (lastblock < firstblock || lastblock == InvalidBlockNumber)
-		return false;
-
-	MemSet(zerobuf, 0, sizeof(zerobuf));
-
-	RelationOpenSmgr(rel);
-
-	/*
-	 * XXX If the extension results in creation of new segment files,
-	 * we have to make sure that each non-last file is correctly filled out to
-	 * RELSEG_SIZE blocks.  This ought to be done inside mdextend, but
-	 * changing the smgr API seems best left for development cycle not late
-	 * beta.  Temporary fix for bug #2737.
-	 */
-#ifndef LET_OS_MANAGE_FILESIZE
-	for (endblock = firstblock | (RELSEG_SIZE - 1);
-		 endblock < lastblock;
-		 endblock += RELSEG_SIZE)
-		smgrextend(rel->rd_smgr, endblock, zerobuf, rel->rd_istemp);
-#endif
-
-	smgrextend(rel->rd_smgr, lastblock, zerobuf, rel->rd_istemp);
-
-	return true;
-}
-
-
 /*
  * _hash_splitbucket -- split 'obucket' into 'obucket' and 'nbucket'
  *
@@ -695,9 +550,9 @@ _hash_splitbucket(Relation rel,
 	TupleDesc	itupdesc = RelationGetDescr(rel);
 
 	/*
-	 * It should be okay to simultaneously write-lock pages from each
-	 * bucket, since no one else can be trying to acquire buffer lock on
-	 * pages of either bucket.
+	 * It should be okay to simultaneously write-lock pages from each bucket,
+	 * since no one else can be trying to acquire buffer lock on pages of
+	 * either bucket.
 	 */
 	oblkno = start_oblkno;
 	nblkno = start_nblkno;
@@ -719,17 +574,17 @@ _hash_splitbucket(Relation rel,
 	nopaque->hasho_filler = HASHO_FILL;
 
 	/*
-	 * Partition the tuples in the old bucket between the old bucket and
-	 * the new bucket, advancing along the old bucket's overflow bucket
-	 * chain and adding overflow pages to the new bucket as needed.
+	 * Partition the tuples in the old bucket between the old bucket and the
+	 * new bucket, advancing along the old bucket's overflow bucket chain and
+	 * adding overflow pages to the new bucket as needed.
 	 */
 	ooffnum = FirstOffsetNumber;
 	omaxoffnum = PageGetMaxOffsetNumber(opage);
 	for (;;)
 	{
 		/*
-		 * at each iteration through this loop, each of these variables
-		 * should be up-to-date: obuf opage oopaque ooffnum omaxoffnum
+		 * at each iteration through this loop, each of these variables should
+		 * be up-to-date: obuf opage oopaque ooffnum omaxoffnum
 		 */
 
 		/* check if we're at the end of the page */
@@ -741,8 +596,8 @@ _hash_splitbucket(Relation rel,
 				break;
 
 			/*
-			 * we ran out of tuples on this particular page, but we have
-			 * more overflow pages; advance to next page.
+			 * we ran out of tuples on this particular page, but we have more
+			 * overflow pages; advance to next page.
 			 */
 			_hash_wrtbuf(rel, obuf);
 
@@ -759,8 +614,7 @@ _hash_splitbucket(Relation rel,
 		 * Re-hash the tuple to determine which bucket it now belongs in.
 		 *
 		 * It is annoying to call the hash function while holding locks, but
-		 * releasing and relocking the page for each tuple is unappealing
-		 * too.
+		 * releasing and relocking the page for each tuple is unappealing too.
 		 */
 		hitem = (HashItem) PageGetItem(opage, PageGetItemId(opage, ooffnum));
 		itup = &(hitem->hash_itup);
@@ -773,9 +627,9 @@ _hash_splitbucket(Relation rel,
 		if (bucket == nbucket)
 		{
 			/*
-			 * insert the tuple into the new bucket.  if it doesn't fit on
-			 * the current page in the new bucket, we must allocate a new
-			 * overflow page and place the tuple on that page instead.
+			 * insert the tuple into the new bucket.  if it doesn't fit on the
+			 * current page in the new bucket, we must allocate a new overflow
+			 * page and place the tuple on that page instead.
 			 */
 			itemsz = IndexTupleDSize(hitem->hash_itup)
 				+ (sizeof(HashItemData) - sizeof(IndexTupleData));
@@ -800,13 +654,13 @@ _hash_splitbucket(Relation rel,
 					 RelationGetRelationName(rel));
 
 			/*
-			 * now delete the tuple from the old bucket.  after this
-			 * section of code, 'ooffnum' will actually point to the
-			 * ItemId to which we would point if we had advanced it before
-			 * the deletion (PageIndexTupleDelete repacks the ItemId
-			 * array).	this also means that 'omaxoffnum' is exactly one
-			 * less than it used to be, so we really can just decrement it
-			 * instead of calling PageGetMaxOffsetNumber.
+			 * now delete the tuple from the old bucket.  after this section
+			 * of code, 'ooffnum' will actually point to the ItemId to which
+			 * we would point if we had advanced it before the deletion
+			 * (PageIndexTupleDelete repacks the ItemId array).  this also
+			 * means that 'omaxoffnum' is exactly one less than it used to be,
+			 * so we really can just decrement it instead of calling
+			 * PageGetMaxOffsetNumber.
 			 */
 			PageIndexTupleDelete(opage, ooffnum);
 			omaxoffnum = OffsetNumberPrev(omaxoffnum);
@@ -814,9 +668,9 @@ _hash_splitbucket(Relation rel,
 		else
 		{
 			/*
-			 * the tuple stays on this page.  we didn't move anything, so
-			 * we didn't delete anything and therefore we don't have to
-			 * change 'omaxoffnum'.
+			 * the tuple stays on this page.  we didn't move anything, so we
+			 * didn't delete anything and therefore we don't have to change
+			 * 'omaxoffnum'.
 			 */
 			Assert(bucket == obucket);
 			ooffnum = OffsetNumberNext(ooffnum);
@@ -824,11 +678,10 @@ _hash_splitbucket(Relation rel,
 	}
 
 	/*
-	 * We're at the end of the old bucket chain, so we're done
-	 * partitioning the tuples.  Before quitting, call _hash_squeezebucket
-	 * to ensure the tuples remaining in the old bucket (including the
-	 * overflow pages) are packed as tightly as possible.  The new bucket
-	 * is already tight.
+	 * We're at the end of the old bucket chain, so we're done partitioning
+	 * the tuples.	Before quitting, call _hash_squeezebucket to ensure the
+	 * tuples remaining in the old bucket (including the overflow pages) are
+	 * packed as tightly as possible.  The new bucket is already tight.
 	 */
 	_hash_wrtbuf(rel, obuf);
 	_hash_wrtbuf(rel, nbuf);
