@@ -3,44 +3,6 @@
  * slru.c
  *		Simple LRU buffering for transaction status logfiles
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
- * Portions Copyright (c) 1994, Regents of the University of California
- *
- * $Header: /cvsroot/pgsql/src/backend/access/transam/slru.c,v 1.7.2.1 2004/02/23 23:03:43 tgl Exp $
- *
- *-------------------------------------------------------------------------
- */
-#include "postgres.h"
-
-#include <fcntl.h>
-#include <sys/stat.h>
-#include <unistd.h>
-
-#include "access/slru.h"
-#include "storage/lwlock.h"
-#include "miscadmin.h"
-
-
-/*
- * Define segment size.  A page is the same BLCKSZ as is used everywhere
- * else in Postgres.  The segment size can be chosen somewhat arbitrarily;
- * we make it 32 pages by default, or 256Kb, i.e. 1M transactions for CLOG
- * or 64K transactions for SUBTRANS.
- *
- * Note: because TransactionIds are 32 bits and wrap around at 0xFFFFFFFF,
- * page numbering also wraps around at 0xFFFFFFFF/xxxx_XACTS_PER_PAGE (where
- * xxxx is CLOG or SUBTRANS, respectively), and segment numbering at
- * 0xFFFFFFFF/xxxx_XACTS_PER_PAGE/SLRU_PAGES_PER_SEGMENT.  We need
- * take no explicit notice of that fact in this module, except when comparing
- * segment and page numbers in SimpleLruTruncate (see PagePrecedes()).
- */
-
-#define SLRU_PAGES_PER_SEGMENT	32
-
-
-/*----------
- * Shared-memory data structures for SLRU control
- *
  * We use a simple least-recently-used scheme to manage a pool of page
  * buffers.  Under ordinary circumstances we expect that write
  * traffic will occur mostly to the latest page (and to the just-prior
@@ -56,10 +18,8 @@
  * that is reading in or writing out a page buffer does not hold the control
  * lock, only the per-buffer lock for the buffer it is working on.
  *
- * To change the page number or state of a buffer, one must normally hold
- * the control lock.  (The sole exception to this rule is that a writer
- * process changes the state from DIRTY to WRITE_IN_PROGRESS while holding
- * only the per-buffer lock.)  If the buffer's state is neither EMPTY nor
+ * To change the page number or state of a buffer, one must hold
+ * the control lock.  If the buffer's state is neither EMPTY nor
  * CLEAN, then there may be processes doing (or waiting to do) I/O on the
  * buffer, so the page number may not be changed, and the only allowed state
  * transition is to change WRITE_IN_PROGRESS to DIRTY after dirtying the page.
@@ -73,55 +33,68 @@
  * the read, while the early marking prevents someone else from trying to
  * read the same page into a different buffer.
  *
- * Note we are assuming that read and write of the state value is atomic,
- * since I/O processes may examine and change the state while not holding
- * the control lock.
- *
  * As with the regular buffer manager, it is possible for another process
  * to re-dirty a page that is currently being written out.	This is handled
  * by setting the page's state from WRITE_IN_PROGRESS to DIRTY.  The writing
  * process must notice this and not mark the page CLEAN when it's done.
- *----------
+ *
+ *
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1994, Regents of the University of California
+ *
+ * $PostgreSQL: pgsql/src/backend/access/transam/slru.c,v 1.29 2005/11/03 00:23:36 tgl Exp $
+ *
+ *-------------------------------------------------------------------------
  */
+#include "postgres.h"
+
+#include <fcntl.h>
+#include <sys/stat.h>
+#include <unistd.h>
+
+#include "access/slru.h"
+#include "access/xlog.h"
+#include "storage/fd.h"
+#include "storage/shmem.h"
+#include "miscadmin.h"
 
-typedef enum
-{
-	SLRU_PAGE_EMPTY,			/* buffer is not in use */
-	SLRU_PAGE_READ_IN_PROGRESS, /* page is being read in */
-	SLRU_PAGE_CLEAN,			/* page is valid and not dirty */
-	SLRU_PAGE_DIRTY,			/* page is valid but needs write */
-	SLRU_PAGE_WRITE_IN_PROGRESS /* page is being written out */
-} SlruPageStatus;
 
 /*
- * Shared-memory state
+ * Define segment size.  A page is the same BLCKSZ as is used everywhere
+ * else in Postgres.  The segment size can be chosen somewhat arbitrarily;
+ * we make it 32 pages by default, or 256Kb, i.e. 1M transactions for CLOG
+ * or 64K transactions for SUBTRANS.
+ *
+ * Note: because TransactionIds are 32 bits and wrap around at 0xFFFFFFFF,
+ * page numbering also wraps around at 0xFFFFFFFF/xxxx_XACTS_PER_PAGE (where
+ * xxxx is CLOG or SUBTRANS, respectively), and segment numbering at
+ * 0xFFFFFFFF/xxxx_XACTS_PER_PAGE/SLRU_PAGES_PER_SEGMENT.  We need
+ * take no explicit notice of that fact in this module, except when comparing
+ * segment and page numbers in SimpleLruTruncate (see PagePrecedes()).
+ *
+ * Note: this file currently assumes that segment file names will be four
+ * hex digits.	This sets a lower bound on the segment size (64K transactions
+ * for 32-bit TransactionIds).
  */
-typedef struct SlruSharedData
-{
-	/*
-	 * Info for each buffer slot.  Page number is undefined when status is
-	 * EMPTY.  lru_count is essentially the number of page switches since
-	 * last use of this page; the page with highest lru_count is the best
-	 * candidate to replace.
-	 */
-	char	   *page_buffer[NUM_CLOG_BUFFERS];
-	SlruPageStatus page_status[NUM_CLOG_BUFFERS];
-	int			page_number[NUM_CLOG_BUFFERS];
-	unsigned int page_lru_count[NUM_CLOG_BUFFERS];
-
-	/*
-	 * latest_page_number is the page number of the current end of the
-	 * CLOG; this is not critical data, since we use it only to avoid
-	 * swapping out the latest page.
-	 */
-	int			latest_page_number;
-} SlruSharedData;
-typedef SlruSharedData *SlruShared;
-
+#define SLRU_PAGES_PER_SEGMENT	32
 
 #define SlruFileName(ctl, path, seg) \
 	snprintf(path, MAXPGPATH, "%s/%04X", (ctl)->Dir, seg)
 
+/*
+ * During SimpleLruFlush(), we will usually not need to write/fsync more
+ * than one or two physical files, but we may need to write several pages
+ * per file.  We can consolidate the I/O requests by leaving files open
+ * until control returns to SimpleLruFlush().  This data structure remembers
+ * which files are open.
+ */
+typedef struct SlruFlushData
+{
+	int			num_files;		/* # files actually open */
+	int			fd[NUM_SLRU_BUFFERS];	/* their FD's */
+	int			segno[NUM_SLRU_BUFFERS];		/* their log seg#s */
+} SlruFlushData;
+
 /*
  * Macro to mark a buffer slot "most recently used".
  */
@@ -129,7 +102,7 @@ typedef SlruSharedData *SlruShared;
 	do { \
 		if ((shared)->page_lru_count[slotno] != 0) { \
 			int		iilru; \
-			for (iilru = 0; iilru < NUM_CLOG_BUFFERS; iilru++) \
+			for (iilru = 0; iilru < NUM_SLRU_BUFFERS; iilru++) \
 				(shared)->page_lru_count[iilru]++; \
 			(shared)->page_lru_count[slotno] = 0; \
 		} \
@@ -142,82 +115,62 @@ typedef enum
 	SLRU_CREATE_FAILED,
 	SLRU_SEEK_FAILED,
 	SLRU_READ_FAILED,
-	SLRU_WRITE_FAILED
+	SLRU_WRITE_FAILED,
+	SLRU_FSYNC_FAILED,
+	SLRU_CLOSE_FAILED
 } SlruErrorCause;
+
 static SlruErrorCause slru_errcause;
 static int	slru_errno;
 
 
 static bool SlruPhysicalReadPage(SlruCtl ctl, int pageno, int slotno);
-static bool SlruPhysicalWritePage(SlruCtl ctl, int pageno, int slotno);
+static bool SlruPhysicalWritePage(SlruCtl ctl, int pageno, int slotno,
+					  SlruFlush fdata);
 static void SlruReportIOError(SlruCtl ctl, int pageno, TransactionId xid);
 static int	SlruSelectLRUPage(SlruCtl ctl, int pageno);
-static bool SlruScanDirectory(SlruCtl ctl, int cutoffPage, bool doDeletions);
 
 
 /*
  * Initialization of shared memory
  */
 
-int
+Size
 SimpleLruShmemSize(void)
 {
-	return MAXALIGN(sizeof(SlruSharedData)) + BLCKSZ * NUM_CLOG_BUFFERS
-#ifdef EXEC_BACKEND
-		+ MAXALIGN(sizeof(SlruLockData))
-#endif
-		;
+	/* we assume NUM_SLRU_BUFFERS isn't so large as to risk overflow */
+	return BUFFERALIGN(sizeof(SlruSharedData)) + BLCKSZ * NUM_SLRU_BUFFERS;
 }
 
 void
-SimpleLruInit(SlruCtl ctl, const char *name, const char *subdir)
+SimpleLruInit(SlruCtl ctl, const char *name,
+			  LWLockId ctllock, const char *subdir)
 {
-	bool		found;
-	char	   *ptr;
 	SlruShared	shared;
-	SlruLock	locks;
-
-	ptr = ShmemInitStruct(name, SimpleLruShmemSize(), &found);
-	shared = (SlruShared) ptr;
-
-#ifdef EXEC_BACKEND
-
-	/*
-	 * Locks are in shared memory
-	 */
-	locks = (SlruLock) (ptr + MAXALIGN(sizeof(SlruSharedData)) +
-						BLCKSZ * NUM_CLOG_BUFFERS);
-#else
-
-	/*
-	 * Locks are in private memory
-	 */
-	Assert(!IsUnderPostmaster);
-	locks = malloc(sizeof(SlruLockData));
-	Assert(locks);
-#endif
+	bool		found;
 
+	shared = (SlruShared) ShmemInitStruct(name, SimpleLruShmemSize(), &found);
 
 	if (!IsUnderPostmaster)
-		/* Initialize locks and shared memory area */
 	{
+		/* Initialize locks and shared memory area */
 		char	   *bufptr;
 		int			slotno;
 
 		Assert(!found);
 
-		locks->ControlLock = LWLockAssign();
-
 		memset(shared, 0, sizeof(SlruSharedData));
 
-		bufptr = (char *) shared + MAXALIGN(sizeof(SlruSharedData));
+		shared->ControlLock = ctllock;
 
-		for (slotno = 0; slotno < NUM_CLOG_BUFFERS; slotno++)
+		bufptr = (char *) shared + BUFFERALIGN(sizeof(SlruSharedData));
+
+		for (slotno = 0; slotno < NUM_SLRU_BUFFERS; slotno++)
 		{
-			locks->BufferLocks[slotno] = LWLockAssign();
 			shared->page_buffer[slotno] = bufptr;
 			shared->page_status[slotno] = SLRU_PAGE_EMPTY;
 			shared->page_lru_count[slotno] = 1;
+			shared->buffer_locks[slotno] = LWLockAssign();
 			bufptr += BLCKSZ;
 		}
 
@@ -226,13 +179,13 @@ SimpleLruInit(SlruCtl ctl, const char *name, const char *subdir)
 	else
 		Assert(found);
 
-
-	ctl->locks = locks;
+	/*
+	 * Initialize the unshared control struct, including directory path. We
+	 * assume caller set PagePrecedes.
+	 */
 	ctl->shared = shared;
-
-
-	/* Init directory path */
-	snprintf(ctl->Dir, MAXPGPATH, "%s/%s", DataDir, subdir);
+	ctl->do_fsync = true;		/* default behavior */
+	StrNCpy(ctl->Dir, subdir, sizeof(ctl->Dir));
 }
 
 /*
@@ -246,8 +199,8 @@ SimpleLruInit(SlruCtl ctl, const char *name, const char *subdir)
 int
 SimpleLruZeroPage(SlruCtl ctl, int pageno)
 {
+	SlruShared	shared = ctl->shared;
 	int			slotno;
-	SlruShared	shared = (SlruShared) ctl->shared;
 
 	/* Find a suitable buffer slot for the page */
 	slotno = SlruSelectLRUPage(ctl, pageno);
@@ -276,16 +229,15 @@ SimpleLruZeroPage(SlruCtl ctl, int pageno)
  * The passed-in xid is used only for error reporting, and may be
  * InvalidTransactionId if no specific xid is associated with the action.
  *
- * Return value is the shared-buffer address of the page.
+ * Return value is the shared-buffer slot number now holding the page.
  * The buffer's LRU access info is updated.
- * If forwrite is true, the buffer is marked as dirty.
  *
  * Control lock must be held at entry, and will be held at exit.
  */
-char *
-SimpleLruReadPage(SlruCtl ctl, int pageno, TransactionId xid, bool forwrite)
+int
+SimpleLruReadPage(SlruCtl ctl, int pageno, TransactionId xid)
 {
-	SlruShared	shared = (SlruShared) ctl->shared;
+	SlruShared	shared = ctl->shared;
 
 	/* Outer loop handles restart if we lose the buffer to someone else */
 	for (;;)
@@ -305,9 +257,7 @@ SimpleLruReadPage(SlruCtl ctl, int pageno, TransactionId xid, bool forwrite)
 			{
 				/* otherwise, it's ready to use */
 				SlruRecentlyUsed(shared, slotno);
-				if (forwrite)
-					shared->page_status[slotno] = SLRU_PAGE_DIRTY;
-				return shared->page_buffer[slotno];
+				return slotno;
 			}
 		}
 		else
@@ -327,9 +277,19 @@ SimpleLruReadPage(SlruCtl ctl, int pageno, TransactionId xid, bool forwrite)
 		 */
 		SlruRecentlyUsed(shared, slotno);
 
-		/* Release shared lock, grab per-buffer lock instead */
-		LWLockRelease(ctl->locks->ControlLock);
-		LWLockAcquire(ctl->locks->BufferLocks[slotno], LW_EXCLUSIVE);
+		/*
+		 * We must grab the per-buffer lock to do I/O.  To avoid deadlock,
+		 * must release ControlLock while waiting for per-buffer lock.
+		 * Fortunately, most of the time the per-buffer lock shouldn't be
+		 * already held, so we can do this:
+		 */
+		if (!LWLockConditionalAcquire(shared->buffer_locks[slotno],
+									  LW_EXCLUSIVE))
+		{
+			LWLockRelease(shared->ControlLock);
+			LWLockAcquire(shared->buffer_locks[slotno], LW_EXCLUSIVE);
+			LWLockAcquire(shared->ControlLock, LW_EXCLUSIVE);
+		}
 
 		/*
 		 * Check to see if someone else already did the read, or took the
@@ -338,32 +298,31 @@ SimpleLruReadPage(SlruCtl ctl, int pageno, TransactionId xid, bool forwrite)
 		if (shared->page_number[slotno] != pageno ||
 			shared->page_status[slotno] != SLRU_PAGE_READ_IN_PROGRESS)
 		{
-			LWLockRelease(ctl->locks->BufferLocks[slotno]);
-			LWLockAcquire(ctl->locks->ControlLock, LW_EXCLUSIVE);
+			LWLockRelease(shared->buffer_locks[slotno]);
 			continue;
 		}
 
-		/* Okay, do the read */
+		/* Okay, release control lock and do the read */
+		LWLockRelease(shared->ControlLock);
+
 		ok = SlruPhysicalReadPage(ctl, pageno, slotno);
 
 		/* Re-acquire shared control lock and update page state */
-		LWLockAcquire(ctl->locks->ControlLock, LW_EXCLUSIVE);
+		LWLockAcquire(shared->ControlLock, LW_EXCLUSIVE);
 
 		Assert(shared->page_number[slotno] == pageno &&
 			   shared->page_status[slotno] == SLRU_PAGE_READ_IN_PROGRESS);
 
 		shared->page_status[slotno] = ok ? SLRU_PAGE_CLEAN : SLRU_PAGE_EMPTY;
 
-		LWLockRelease(ctl->locks->BufferLocks[slotno]);
+		LWLockRelease(shared->buffer_locks[slotno]);
 
 		/* Now it's okay to ereport if we failed */
 		if (!ok)
 			SlruReportIOError(ctl, pageno, xid);
 
 		SlruRecentlyUsed(shared, slotno);
-		if (forwrite)
-			shared->page_status[slotno] = SLRU_PAGE_DIRTY;
-		return shared->page_buffer[slotno];
+		return slotno;
 	}
 }
 
@@ -379,11 +338,11 @@ SimpleLruReadPage(SlruCtl ctl, int pageno, TransactionId xid, bool forwrite)
  * Control lock must be held at entry, and will be held at exit.
  */
 void
-SimpleLruWritePage(SlruCtl ctl, int slotno)
+SimpleLruWritePage(SlruCtl ctl, int slotno, SlruFlush fdata)
 {
+	SlruShared	shared = ctl->shared;
 	int			pageno;
 	bool		ok;
-	SlruShared	shared = (SlruShared) ctl->shared;
 
 	/* Do nothing if page does not need writing */
 	if (shared->page_status[slotno] != SLRU_PAGE_DIRTY &&
@@ -392,45 +351,57 @@ SimpleLruWritePage(SlruCtl ctl, int slotno)
 
 	pageno = shared->page_number[slotno];
 
-	/* Release shared lock, grab per-buffer lock instead */
-	LWLockRelease(ctl->locks->ControlLock);
-	LWLockAcquire(ctl->locks->BufferLocks[slotno], LW_EXCLUSIVE);
+	/*
+	 * We must grab the per-buffer lock to do I/O.  To avoid deadlock,
+	 * must release ControlLock while waiting for per-buffer lock.
+	 * Fortunately, most of the time the per-buffer lock shouldn't be
+	 * already held, so we can do this:
+	 */
+	if (!LWLockConditionalAcquire(shared->buffer_locks[slotno],
+								  LW_EXCLUSIVE))
+	{
+		LWLockRelease(shared->ControlLock);
+		LWLockAcquire(shared->buffer_locks[slotno], LW_EXCLUSIVE);
+		LWLockAcquire(shared->ControlLock, LW_EXCLUSIVE);
+	}
 
 	/*
-	 * Check to see if someone else already did the write, or took the
-	 * buffer away from us.  If so, do nothing.  NOTE: we really should
-	 * never see WRITE_IN_PROGRESS here, since that state should only
-	 * occur while the writer is holding the buffer lock.  But accept it
-	 * so that we have a recovery path if a writer aborts.
+	 * Check to see if someone else already did the write, or took the buffer
+	 * away from us.  If so, do nothing.  NOTE: we really should never see
+	 * WRITE_IN_PROGRESS here, since that state should only occur while the
+	 * writer is holding the buffer lock.  But accept it so that we have a
+	 * recovery path if a writer aborts.
 	 */
 	if (shared->page_number[slotno] != pageno ||
 		(shared->page_status[slotno] != SLRU_PAGE_DIRTY &&
 		 shared->page_status[slotno] != SLRU_PAGE_WRITE_IN_PROGRESS))
 	{
-		LWLockRelease(ctl->locks->BufferLocks[slotno]);
-		LWLockAcquire(ctl->locks->ControlLock, LW_EXCLUSIVE);
+		LWLockRelease(shared->buffer_locks[slotno]);
 		return;
 	}
 
 	/*
 	 * Mark the slot write-busy.  After this point, a transaction status
-	 * update on this page will mark it dirty again.  NB: we are assuming
-	 * that read/write of the page status field is atomic, since we change
-	 * the state while not holding control lock.  However, we cannot set
-	 * this state any sooner, or we'd possibly fool a previous writer into
-	 * thinking he's successfully dumped the page when he hasn't.
-	 * (Scenario: other writer starts, page is redirtied, we come along
-	 * and set WRITE_IN_PROGRESS again, other writer completes and sets
-	 * CLEAN because redirty info has been lost, then we think it's clean
-	 * too.)
+	 * update on this page will mark it dirty again.
 	 */
 	shared->page_status[slotno] = SLRU_PAGE_WRITE_IN_PROGRESS;
 
-	/* Okay, do the write */
-	ok = SlruPhysicalWritePage(ctl, pageno, slotno);
+	/* Okay, release the control lock and do the write */
+	LWLockRelease(shared->ControlLock);
+
+	ok = SlruPhysicalWritePage(ctl, pageno, slotno, fdata);
+
+	/* If we failed, and we're in a flush, better close the files */
+	if (!ok && fdata)
+	{
+		int			i;
+
+		for (i = 0; i < fdata->num_files; i++)
+			close(fdata->fd[i]);
+	}
 
 	/* Re-acquire shared control lock and update page state */
-	LWLockAcquire(ctl->locks->ControlLock, LW_EXCLUSIVE);
+	LWLockAcquire(shared->ControlLock, LW_EXCLUSIVE);
 
 	Assert(shared->page_number[slotno] == pageno &&
 		   (shared->page_status[slotno] == SLRU_PAGE_WRITE_IN_PROGRESS ||
@@ -440,7 +411,7 @@ SimpleLruWritePage(SlruCtl ctl, int slotno)
 	if (shared->page_status[slotno] == SLRU_PAGE_WRITE_IN_PROGRESS)
 		shared->page_status[slotno] = ok ? SLRU_PAGE_CLEAN : SLRU_PAGE_DIRTY;
 
-	LWLockRelease(ctl->locks->BufferLocks[slotno]);
+	LWLockRelease(shared->buffer_locks[slotno]);
 
 	/* Now it's okay to ereport if we failed */
 	if (!ok)
@@ -460,7 +431,7 @@ SimpleLruWritePage(SlruCtl ctl, int slotno)
 static bool
 SlruPhysicalReadPage(SlruCtl ctl, int pageno, int slotno)
 {
-	SlruShared	shared = (SlruShared) ctl->shared;
+	SlruShared	shared = ctl->shared;
 	int			segno = pageno / SLRU_PAGES_PER_SEGMENT;
 	int			rpageno = pageno % SLRU_PAGES_PER_SEGMENT;
 	int			offset = rpageno * BLCKSZ;
@@ -473,8 +444,8 @@ SlruPhysicalReadPage(SlruCtl ctl, int pageno, int slotno)
 	 * In a crash-and-restart situation, it's possible for us to receive
 	 * commands to set the commit status of transactions whose bits are in
 	 * already-truncated segments of the commit log (see notes in
-	 * SlruPhysicalWritePage).	Hence, if we are InRecovery, allow the
-	 * case where the file doesn't exist, and return zeroes instead.
+	 * SlruPhysicalWritePage).	Hence, if we are InRecovery, allow the case
+	 * where the file doesn't exist, and return zeroes instead.
 	 */
 	fd = BasicOpenFile(path, O_RDWR | PG_BINARY, S_IRUSR | S_IWUSR);
 	if (fd < 0)
@@ -497,6 +468,7 @@ SlruPhysicalReadPage(SlruCtl ctl, int pageno, int slotno)
 	{
 		slru_errcause = SLRU_SEEK_FAILED;
 		slru_errno = errno;
+		close(fd);
 		return false;
 	}
 
@@ -505,10 +477,17 @@ SlruPhysicalReadPage(SlruCtl ctl, int pageno, int slotno)
 	{
 		slru_errcause = SLRU_READ_FAILED;
 		slru_errno = errno;
+		close(fd);
+		return false;
+	}
+
+	if (close(fd))
+	{
+		slru_errcause = SLRU_CLOSE_FAILED;
+		slru_errno = errno;
 		return false;
 	}
 
-	close(fd);
 	return true;
 }
 
@@ -520,50 +499,80 @@ SlruPhysicalReadPage(SlruCtl ctl, int pageno, int slotno)
  * info in static variables to let SlruReportIOError make the report.
  *
  * For now, assume it's not worth keeping a file pointer open across
- * read/write operations.  We could cache one virtual file pointer ...
+ * independent read/write operations.  We do batch operations during
+ * SimpleLruFlush, though.
+ *
+ * fdata is NULL for a standalone write, pointer to open-file info during
+ * SimpleLruFlush.
  */
 static bool
-SlruPhysicalWritePage(SlruCtl ctl, int pageno, int slotno)
+SlruPhysicalWritePage(SlruCtl ctl, int pageno, int slotno, SlruFlush fdata)
 {
-	SlruShared	shared = (SlruShared) ctl->shared;
+	SlruShared	shared = ctl->shared;
 	int			segno = pageno / SLRU_PAGES_PER_SEGMENT;
 	int			rpageno = pageno % SLRU_PAGES_PER_SEGMENT;
 	int			offset = rpageno * BLCKSZ;
 	char		path[MAXPGPATH];
-	int			fd;
-
-	SlruFileName(ctl, path, segno);
+	int			fd = -1;
 
 	/*
-	 * If the file doesn't already exist, we should create it.  It is
-	 * possible for this to need to happen when writing a page that's not
-	 * first in its segment; we assume the OS can cope with that.  (Note:
-	 * it might seem that it'd be okay to create files only when
-	 * SimpleLruZeroPage is called for the first page of a segment.
-	 * However, if after a crash and restart the REDO logic elects to
-	 * replay the log from a checkpoint before the latest one, then it's
-	 * possible that we will get commands to set transaction status of
-	 * transactions that have already been truncated from the commit log.
-	 * Easiest way to deal with that is to accept references to
-	 * nonexistent files here and in SlruPhysicalReadPage.)
+	 * During a Flush, we may already have the desired file open.
 	 */
-	fd = BasicOpenFile(path, O_RDWR | PG_BINARY, S_IRUSR | S_IWUSR);
+	if (fdata)
+	{
+		int			i;
+
+		for (i = 0; i < fdata->num_files; i++)
+		{
+			if (fdata->segno[i] == segno)
+			{
+				fd = fdata->fd[i];
+				break;
+			}
+		}
+	}
+
 	if (fd < 0)
 	{
-		if (errno != ENOENT)
-		{
-			slru_errcause = SLRU_OPEN_FAILED;
-			slru_errno = errno;
-			return false;
-		}
-
-		fd = BasicOpenFile(path, O_RDWR | O_CREAT | O_EXCL | PG_BINARY,
-						   S_IRUSR | S_IWUSR);
+		/*
+		 * If the file doesn't already exist, we should create it.  It is
+		 * possible for this to need to happen when writing a page that's not
+		 * first in its segment; we assume the OS can cope with that. (Note:
+		 * it might seem that it'd be okay to create files only when
+		 * SimpleLruZeroPage is called for the first page of a segment.
+		 * However, if after a crash and restart the REDO logic elects to
+		 * replay the log from a checkpoint before the latest one, then it's
+		 * possible that we will get commands to set transaction status of
+		 * transactions that have already been truncated from the commit log.
+		 * Easiest way to deal with that is to accept references to
+		 * nonexistent files here and in SlruPhysicalReadPage.)
+		 */
+		SlruFileName(ctl, path, segno);
+		fd = BasicOpenFile(path, O_RDWR | PG_BINARY, S_IRUSR | S_IWUSR);
 		if (fd < 0)
 		{
-			slru_errcause = SLRU_CREATE_FAILED;
-			slru_errno = errno;
-			return false;
+			if (errno != ENOENT)
+			{
+				slru_errcause = SLRU_OPEN_FAILED;
+				slru_errno = errno;
+				return false;
+			}
+
+			fd = BasicOpenFile(path, O_RDWR | O_CREAT | O_EXCL | PG_BINARY,
+							   S_IRUSR | S_IWUSR);
+			if (fd < 0)
+			{
+				slru_errcause = SLRU_CREATE_FAILED;
+				slru_errno = errno;
+				return false;
+			}
+		}
+
+		if (fdata)
+		{
+			fdata->fd[fdata->num_files] = fd;
+			fdata->segno[fdata->num_files] = segno;
+			fdata->num_files++;
 		}
 	}
 
@@ -571,6 +580,8 @@ SlruPhysicalWritePage(SlruCtl ctl, int pageno, int slotno)
 	{
 		slru_errcause = SLRU_SEEK_FAILED;
 		slru_errno = errno;
+		if (!fdata)
+			close(fd);
 		return false;
 	}
 
@@ -582,10 +593,33 @@ SlruPhysicalWritePage(SlruCtl ctl, int pageno, int slotno)
 			errno = ENOSPC;
 		slru_errcause = SLRU_WRITE_FAILED;
 		slru_errno = errno;
+		if (!fdata)
+			close(fd);
 		return false;
 	}
 
-	close(fd);
+	/*
+	 * If not part of Flush, need to fsync now.  We assume this happens
+	 * infrequently enough that it's not a performance issue.
+	 */
+	if (!fdata)
+	{
+		if (ctl->do_fsync && pg_fsync(fd))
+		{
+			slru_errcause = SLRU_FSYNC_FAILED;
+			slru_errno = errno;
+			close(fd);
+			return false;
+		}
+
+		if (close(fd))
+		{
+			slru_errcause = SLRU_CLOSE_FAILED;
+			slru_errno = errno;
+			return false;
+		}
+	}
+
 	return true;
 }
 
@@ -608,37 +642,51 @@ SlruReportIOError(SlruCtl ctl, int pageno, TransactionId xid)
 		case SLRU_OPEN_FAILED:
 			ereport(ERROR,
 					(errcode_for_file_access(),
-				errmsg("could not access status of transaction %u", xid),
+					 errmsg("could not access status of transaction %u", xid),
 					 errdetail("could not open file \"%s\": %m",
 							   path)));
 			break;
 		case SLRU_CREATE_FAILED:
 			ereport(ERROR,
 					(errcode_for_file_access(),
-				errmsg("could not access status of transaction %u", xid),
+					 errmsg("could not access status of transaction %u", xid),
 					 errdetail("could not create file \"%s\": %m",
 							   path)));
 			break;
 		case SLRU_SEEK_FAILED:
 			ereport(ERROR,
 					(errcode_for_file_access(),
-				errmsg("could not access status of transaction %u", xid),
+					 errmsg("could not access status of transaction %u", xid),
 				  errdetail("could not seek in file \"%s\" to offset %u: %m",
 							path, offset)));
 			break;
 		case SLRU_READ_FAILED:
 			ereport(ERROR,
 					(errcode_for_file_access(),
-				errmsg("could not access status of transaction %u", xid),
-				   errdetail("could not read from file \"%s\" at offset %u: %m",
-							 path, offset)));
+					 errmsg("could not access status of transaction %u", xid),
+				errdetail("could not read from file \"%s\" at offset %u: %m",
+						  path, offset)));
 			break;
 		case SLRU_WRITE_FAILED:
 			ereport(ERROR,
 					(errcode_for_file_access(),
-				errmsg("could not access status of transaction %u", xid),
-				  errdetail("could not write to file \"%s\" at offset %u: %m",
-							path, offset)));
+					 errmsg("could not access status of transaction %u", xid),
+				 errdetail("could not write to file \"%s\" at offset %u: %m",
+						   path, offset)));
+			break;
+		case SLRU_FSYNC_FAILED:
+			ereport(ERROR,
+					(errcode_for_file_access(),
+					 errmsg("could not access status of transaction %u", xid),
+					 errdetail("could not fsync file \"%s\": %m",
+							   path)));
+			break;
+		case SLRU_CLOSE_FAILED:
+			ereport(ERROR,
+					(errcode_for_file_access(),
+					 errmsg("could not access status of transaction %u", xid),
+					 errdetail("could not close file \"%s\": %m",
+							   path)));
 			break;
 		default:
 			/* can't get here, we trust */
@@ -664,7 +712,7 @@ SlruReportIOError(SlruCtl ctl, int pageno, TransactionId xid)
 static int
 SlruSelectLRUPage(SlruCtl ctl, int pageno)
 {
-	SlruShared	shared = (SlruShared) ctl->shared;
+	SlruShared	shared = ctl->shared;
 
 	/* Outer loop handles restart after I/O */
 	for (;;)
@@ -674,7 +722,7 @@ SlruSelectLRUPage(SlruCtl ctl, int pageno)
 		unsigned int bestcount = 0;
 
 		/* See if page already has a buffer assigned */
-		for (slotno = 0; slotno < NUM_CLOG_BUFFERS; slotno++)
+		for (slotno = 0; slotno < NUM_SLRU_BUFFERS; slotno++)
 		{
 			if (shared->page_number[slotno] == pageno &&
 				shared->page_status[slotno] != SLRU_PAGE_EMPTY)
@@ -682,10 +730,10 @@ SlruSelectLRUPage(SlruCtl ctl, int pageno)
 		}
 
 		/*
-		 * If we find any EMPTY slot, just select that one. Else locate
-		 * the least-recently-used slot that isn't the latest page.
+		 * If we find any EMPTY slot, just select that one. Else locate the
+		 * least-recently-used slot that isn't the latest page.
 		 */
-		for (slotno = 0; slotno < NUM_CLOG_BUFFERS; slotno++)
+		for (slotno = 0; slotno < NUM_SLRU_BUFFERS; slotno++)
 		{
 			if (shared->page_status[slotno] == SLRU_PAGE_EMPTY)
 				return slotno;
@@ -704,123 +752,131 @@ SlruSelectLRUPage(SlruCtl ctl, int pageno)
 			return bestslot;
 
 		/*
-		 * We need to do I/O.  Normal case is that we have to write it
-		 * out, but it's possible in the worst case to have selected a
-		 * read-busy page.	In that case we use SimpleLruReadPage to wait
-		 * for the read to complete.
+		 * We need to do I/O.  Normal case is that we have to write it out,
+		 * but it's possible in the worst case to have selected a read-busy
+		 * page.  In that case we just wait for someone else to complete
+		 * the I/O, which we can do by waiting for the per-buffer lock.
 		 */
 		if (shared->page_status[bestslot] == SLRU_PAGE_READ_IN_PROGRESS)
-			(void) SimpleLruReadPage(ctl, shared->page_number[bestslot],
-									 InvalidTransactionId, false);
+		{
+			LWLockRelease(shared->ControlLock);
+			LWLockAcquire(shared->buffer_locks[bestslot], LW_SHARED);
+			LWLockRelease(shared->buffer_locks[bestslot]);
+			LWLockAcquire(shared->ControlLock, LW_EXCLUSIVE);
+		}
 		else
-			SimpleLruWritePage(ctl, bestslot);
+			SimpleLruWritePage(ctl, bestslot, NULL);
 
 		/*
-		 * Now loop back and try again.  This is the easiest way of
-		 * dealing with corner cases such as the victim page being
-		 * re-dirtied while we wrote it.
+		 * Now loop back and try again.  This is the easiest way of dealing
+		 * with corner cases such as the victim page being re-dirtied while we
+		 * wrote it.
 		 */
 	}
 }
 
 /*
- * This must be called ONCE during postmaster or standalone-backend startup
- */
-void
-SimpleLruSetLatestPage(SlruCtl ctl, int pageno)
-{
-	SlruShared	shared = (SlruShared) ctl->shared;
-
-	shared->latest_page_number = pageno;
-}
-
-/*
- * This is called during checkpoint and postmaster/standalone-backend shutdown
+ * Flush dirty pages to disk during checkpoint or database shutdown
  */
 void
 SimpleLruFlush(SlruCtl ctl, bool checkpoint)
 {
-#ifdef USE_ASSERT_CHECKING		/* only used in Assert() */
-	SlruShared	shared = (SlruShared) ctl->shared;
-#endif
+	SlruShared	shared = ctl->shared;
+	SlruFlushData fdata;
 	int			slotno;
+	int			pageno = 0;
+	int			i;
+	bool		ok;
 
-	LWLockAcquire(ctl->locks->ControlLock, LW_EXCLUSIVE);
+	/*
+	 * Find and write dirty pages
+	 */
+	fdata.num_files = 0;
 
-	for (slotno = 0; slotno < NUM_CLOG_BUFFERS; slotno++)
+	LWLockAcquire(shared->ControlLock, LW_EXCLUSIVE);
+
+	for (slotno = 0; slotno < NUM_SLRU_BUFFERS; slotno++)
 	{
-		SimpleLruWritePage(ctl, slotno);
+		SimpleLruWritePage(ctl, slotno, &fdata);
 
 		/*
-		 * When called during a checkpoint, we cannot assert that the slot
-		 * is clean now, since another process might have re-dirtied it
-		 * already.  That's okay.
+		 * When called during a checkpoint, we cannot assert that the slot is
+		 * clean now, since another process might have re-dirtied it already.
+		 * That's okay.
 		 */
 		Assert(checkpoint ||
 			   shared->page_status[slotno] == SLRU_PAGE_EMPTY ||
 			   shared->page_status[slotno] == SLRU_PAGE_CLEAN);
 	}
 
-	LWLockRelease(ctl->locks->ControlLock);
+	LWLockRelease(shared->ControlLock);
+
+	/*
+	 * Now fsync and close any files that were open
+	 */
+	ok = true;
+	for (i = 0; i < fdata.num_files; i++)
+	{
+		if (ctl->do_fsync && pg_fsync(fdata.fd[i]))
+		{
+			slru_errcause = SLRU_FSYNC_FAILED;
+			slru_errno = errno;
+			pageno = fdata.segno[i] * SLRU_PAGES_PER_SEGMENT;
+			ok = false;
+		}
+
+		if (close(fdata.fd[i]))
+		{
+			slru_errcause = SLRU_CLOSE_FAILED;
+			slru_errno = errno;
+			pageno = fdata.segno[i] * SLRU_PAGES_PER_SEGMENT;
+			ok = false;
+		}
+	}
+	if (!ok)
+		SlruReportIOError(ctl, pageno, InvalidTransactionId);
 }
 
 /*
  * Remove all segments before the one holding the passed page number
- *
- * When this is called, we know that the database logically contains no
- * reference to transaction IDs older than oldestXact.	However, we must
- * not remove any segment until we have performed a checkpoint, to ensure
- * that no such references remain on disk either; else a crash just after
- * the truncation might leave us with a problem.  Since CLOG segments hold
- * a large number of transactions, the opportunity to actually remove a
- * segment is fairly rare, and so it seems best not to do the checkpoint
- * unless we have confirmed that there is a removable segment.	Therefore
- * we issue the checkpoint command here, not in higher-level code as might
- * seem cleaner.
  */
 void
 SimpleLruTruncate(SlruCtl ctl, int cutoffPage)
 {
+	SlruShared	shared = ctl->shared;
 	int			slotno;
-	SlruShared	shared = (SlruShared) ctl->shared;
 
 	/*
 	 * The cutoff point is the start of the segment containing cutoffPage.
 	 */
 	cutoffPage -= cutoffPage % SLRU_PAGES_PER_SEGMENT;
 
-	if (!SlruScanDirectory(ctl, cutoffPage, false))
-		return;					/* nothing to remove */
-
-	/* Perform a forced CHECKPOINT */
-	CreateCheckPoint(false, true);
-
 	/*
-	 * Scan shared memory and remove any pages preceding the cutoff page,
-	 * to ensure we won't rewrite them later.  (Any dirty pages should
-	 * have been flushed already during the checkpoint, we're just being
-	 * extra careful here.)
+	 * Scan shared memory and remove any pages preceding the cutoff page, to
+	 * ensure we won't rewrite them later.  (Since this is normally called in
+	 * or just after a checkpoint, any dirty pages should have been flushed
+	 * already ... we're just being extra careful here.)
 	 */
-	LWLockAcquire(ctl->locks->ControlLock, LW_EXCLUSIVE);
+	LWLockAcquire(shared->ControlLock, LW_EXCLUSIVE);
 
 restart:;
 
 	/*
 	 * While we are holding the lock, make an important safety check: the
-	 * planned cutoff point must be <= the current endpoint page.
-	 * Otherwise we have already wrapped around, and proceeding with the
-	 * truncation would risk removing the current segment.
+	 * planned cutoff point must be <= the current endpoint page. Otherwise we
+	 * have already wrapped around, and proceeding with the truncation would
+	 * risk removing the current segment.
 	 */
 	if (ctl->PagePrecedes(shared->latest_page_number, cutoffPage))
 	{
-		LWLockRelease(ctl->locks->ControlLock);
+		LWLockRelease(shared->ControlLock);
 		ereport(LOG,
-				(errmsg("could not truncate directory \"%s\": apparent wraparound",
-						ctl->Dir)));
+		  (errmsg("could not truncate directory \"%s\": apparent wraparound",
+				  ctl->Dir)));
 		return;
 	}
 
-	for (slotno = 0; slotno < NUM_CLOG_BUFFERS; slotno++)
+	for (slotno = 0; slotno < NUM_SLRU_BUFFERS; slotno++)
 	{
 		if (shared->page_status[slotno] == SLRU_PAGE_EMPTY)
 			continue;
@@ -837,30 +893,36 @@ restart:;
 		}
 
 		/*
-		 * Hmm, we have (or may have) I/O operations acting on the page,
-		 * so we've got to wait for them to finish and then start again.
-		 * This is the same logic as in SlruSelectLRUPage.
+		 * Hmm, we have (or may have) I/O operations acting on the page, so
+		 * we've got to wait for them to finish and then start again. This is
+		 * the same logic as in SlruSelectLRUPage.
 		 */
 		if (shared->page_status[slotno] == SLRU_PAGE_READ_IN_PROGRESS)
-			(void) SimpleLruReadPage(ctl, shared->page_number[slotno],
-									 InvalidTransactionId, false);
+		{
+			LWLockRelease(shared->ControlLock);
+			LWLockAcquire(shared->buffer_locks[slotno], LW_SHARED);
+			LWLockRelease(shared->buffer_locks[slotno]);
+			LWLockAcquire(shared->ControlLock, LW_EXCLUSIVE);
+		}
 		else
-			SimpleLruWritePage(ctl, slotno);
+			SimpleLruWritePage(ctl, slotno, NULL);
 		goto restart;
 	}
 
-	LWLockRelease(ctl->locks->ControlLock);
+	LWLockRelease(shared->ControlLock);
 
 	/* Now we can remove the old segment(s) */
 	(void) SlruScanDirectory(ctl, cutoffPage, true);
 }
 
 /*
- * SlruTruncate subroutine: scan directory for removable segments.
+ * SimpleLruTruncate subroutine: scan directory for removable segments.
  * Actually remove them iff doDeletions is true.  Return TRUE iff any
  * removable segments were found.  Note: no locking is needed.
+ *
+ * This can be called directly from clog.c, for reasons explained there.
  */
-static bool
+bool
 SlruScanDirectory(SlruCtl ctl, int cutoffPage, bool doDeletions)
 {
 	bool		found = false;
@@ -870,14 +932,15 @@ SlruScanDirectory(SlruCtl ctl, int cutoffPage, bool doDeletions)
 	int			segpage;
 	char		path[MAXPGPATH];
 
-	cldir = AllocateDir(ctl->Dir);
-	if (cldir == NULL)
-		ereport(ERROR,
-				(errcode_for_file_access(),
-			   errmsg("could not open directory \"%s\": %m", ctl->Dir)));
+	/*
+	 * The cutoff point is the start of the segment containing cutoffPage.
+	 * (This is redundant when called from SimpleLruTruncate, but not when
+	 * called directly from clog.c.)
+	 */
+	cutoffPage -= cutoffPage % SLRU_PAGES_PER_SEGMENT;
 
-	errno = 0;
-	while ((clde = readdir(cldir)) != NULL)
+	cldir = AllocateDir(ctl->Dir);
+	while ((clde = ReadDir(cldir, ctl->Dir)) != NULL)
 	{
 		if (strlen(clde->d_name) == 4 &&
 			strspn(clde->d_name, "0123456789ABCDEF") == 4)
@@ -889,20 +952,14 @@ SlruScanDirectory(SlruCtl ctl, int cutoffPage, bool doDeletions)
 				found = true;
 				if (doDeletions)
 				{
-					ereport(LOG,
-							(errmsg("removing file \"%s/%s\"",
-									ctl->Dir, clde->d_name)));
 					snprintf(path, MAXPGPATH, "%s/%s", ctl->Dir, clde->d_name);
+					ereport(DEBUG2,
+							(errmsg("removing file \"%s\"", path)));
 					unlink(path);
 				}
 			}
 		}
-		errno = 0;
 	}
-	if (errno)
-		ereport(ERROR,
-				(errcode_for_file_access(),
-			   errmsg("could not read directory \"%s\": %m", ctl->Dir)));
 	FreeDir(cldir);
 
 	return found;
