@@ -37,7 +37,7 @@
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/commands/tablespace.c,v 1.15.4.2 2006/03/29 15:16:00 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/commands/tablespace.c,v 1.28 2005/10/15 02:49:15 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -50,7 +50,7 @@
 
 #include "access/heapam.h"
 #include "catalog/catalog.h"
-#include "catalog/catname.h"
+#include "catalog/dependency.h"
 #include "catalog/indexing.h"
 #include "catalog/pg_namespace.h"
 #include "catalog/pg_tablespace.h"
@@ -67,7 +67,7 @@
 
 
 /* GUC variable */
-char   *default_tablespace = NULL;
+char	   *default_tablespace = NULL;
 
 
 static bool remove_tablespace_directories(Oid tablespaceoid, bool redo);
@@ -80,21 +80,24 @@ static void set_short_version(const char *path);
  * object in the tablespace, create the subdirectory.  If the subdirectory
  * already exists, just fall through quietly.
  *
- * isRedo indicates that we are creating an object during WAL replay.
- * In this case we will cope with the possibility of the tablespace
- * directory not being there either --- this could happen if we are
+ * If tablespaces are not supported, this is just a no-op; CREATE DATABASE
+ * is expected to create the default subdirectory for the database.
+ *
+ * isRedo indicates that we are creating an object during WAL replay;
+ * we can skip doing locking in that case (and should do so to avoid
+ * any possible problems with pg_tablespace not being valid).
+ *
+ * Also, when isRedo is true, we will cope with the possibility of the
+ * tablespace not being there either --- this could happen if we are
  * replaying an operation on a table in a subsequently-dropped tablespace.
  * We handle this by making a directory in the place where the tablespace
  * symlink would normally be.  This isn't an exact replay of course, but
  * it's the best we can do given the available information.
- *
- * If tablespaces are not supported, you might think this could be a no-op,
- * but you'd be wrong: we still need it in case we have to re-create a
- * database subdirectory (of $PGDATA/base) during WAL replay.
  */
 void
 TablespaceCreateDbspace(Oid spcNode, Oid dbNode, bool isRedo)
 {
+#ifdef HAVE_SYMLINK
 	struct stat st;
 	char	   *dir;
 
@@ -115,14 +118,20 @@ TablespaceCreateDbspace(Oid spcNode, Oid dbNode, bool isRedo)
 		if (errno == ENOENT)
 		{
 			/*
-			 * Acquire TablespaceCreateLock to ensure that no DROP TABLESPACE
-			 * or TablespaceCreateDbspace is running concurrently.
+			 * Acquire ExclusiveLock on pg_tablespace to ensure that no DROP
+			 * TABLESPACE or TablespaceCreateDbspace is running concurrently.
+			 * Simple reads from pg_tablespace are OK.
 			 */
-			LWLockAcquire(TablespaceCreateLock, LW_EXCLUSIVE);
+			Relation	rel;
+
+			if (!isRedo)
+				rel = heap_open(TableSpaceRelationId, ExclusiveLock);
+			else
+				rel = NULL;
 
 			/*
-			 * Recheck to see if someone created the directory while we
-			 * were waiting for lock.
+			 * Recheck to see if someone created the directory while we were
+			 * waiting for lock.
 			 */
 			if (stat(dir, &st) == 0 && S_ISDIR(st.st_mode))
 			{
@@ -138,26 +147,28 @@ TablespaceCreateDbspace(Oid spcNode, Oid dbNode, bool isRedo)
 					if (errno != ENOENT || !isRedo)
 						ereport(ERROR,
 								(errcode_for_file_access(),
-						  errmsg("could not create directory \"%s\": %m",
-								 dir)));
+							  errmsg("could not create directory \"%s\": %m",
+									 dir)));
 					/* Try to make parent directory too */
 					parentdir = pstrdup(dir);
 					get_parent_directory(parentdir);
 					if (mkdir(parentdir, S_IRWXU) < 0)
 						ereport(ERROR,
 								(errcode_for_file_access(),
-						  errmsg("could not create directory \"%s\": %m",
-								 parentdir)));
+							  errmsg("could not create directory \"%s\": %m",
+									 parentdir)));
 					pfree(parentdir);
 					if (mkdir(dir, S_IRWXU) < 0)
 						ereport(ERROR,
 								(errcode_for_file_access(),
-						  errmsg("could not create directory \"%s\": %m",
-								 dir)));
+							  errmsg("could not create directory \"%s\": %m",
+									 dir)));
 				}
 			}
 
-			LWLockRelease(TablespaceCreateLock);
+			/* OK to drop the exclusive lock */
+			if (!isRedo)
+				heap_close(rel, ExclusiveLock);
 		}
 		else
 		{
@@ -177,6 +188,7 @@ TablespaceCreateDbspace(Oid spcNode, Oid dbNode, bool isRedo)
 	}
 
 	pfree(dir);
+#endif   /* HAVE_SYMLINK */
 }
 
 /*
@@ -197,7 +209,7 @@ CreateTableSpace(CreateTableSpaceStmt *stmt)
 	Oid			tablespaceoid;
 	char	   *location;
 	char	   *linkloc;
-	AclId		ownerid;
+	Oid			ownerId;
 
 	/* validate */
 
@@ -214,12 +226,9 @@ CreateTableSpace(CreateTableSpaceStmt *stmt)
 
 	/* However, the eventual owner of the tablespace need not be */
 	if (stmt->owner)
-	{
-		/* No need to check result, get_usesysid() does that */
-		ownerid = get_usesysid(stmt->owner);
-	}
+		ownerId = get_roleid_checked(stmt->owner);
 	else
-		ownerid = GetUserId();
+		ownerId = GetUserId();
 
 	/* Unix-ify the offered path, and strip any trailing slashes */
 	location = pstrdup(stmt->location);
@@ -229,7 +238,7 @@ CreateTableSpace(CreateTableSpaceStmt *stmt)
 	if (strchr(location, '\''))
 		ereport(ERROR,
 				(errcode(ERRCODE_INVALID_NAME),
-		   errmsg("tablespace location may not contain single quotes")));
+			   errmsg("tablespace location may not contain single quotes")));
 
 	/*
 	 * Allowing relative paths seems risky
@@ -242,9 +251,9 @@ CreateTableSpace(CreateTableSpaceStmt *stmt)
 				 errmsg("tablespace location must be an absolute path")));
 
 	/*
-	 * Check that location isn't too long. Remember that we're going to
-	 * append '/<dboid>/<relid>.<nnn>'	(XXX but do we ever form the whole
-	 * path explicitly?  This may be overly conservative.)
+	 * Check that location isn't too long. Remember that we're going to append
+	 * '/<dboid>/<relid>.<nnn>'  (XXX but do we ever form the whole path
+	 * explicitly?	This may be overly conservative.)
 	 */
 	if (strlen(location) >= (MAXPGPATH - 1 - 10 - 1 - 10 - 1 - 10))
 		ereport(ERROR,
@@ -261,7 +270,7 @@ CreateTableSpace(CreateTableSpaceStmt *stmt)
 				(errcode(ERRCODE_RESERVED_NAME),
 				 errmsg("unacceptable tablespace name \"%s\"",
 						stmt->tablespacename),
-				 errdetail("The prefix \"pg_\" is reserved for system tablespaces.")));
+		errdetail("The prefix \"pg_\" is reserved for system tablespaces.")));
 
 	/*
 	 * Check that there is no other tablespace by this name.  (The unique
@@ -275,18 +284,18 @@ CreateTableSpace(CreateTableSpaceStmt *stmt)
 						stmt->tablespacename)));
 
 	/*
-	 * Insert tuple into pg_tablespace.  The purpose of doing this first
-	 * is to lock the proposed tablename against other would-be creators.
-	 * The insertion will roll back if we find problems below.
+	 * Insert tuple into pg_tablespace.  The purpose of doing this first is to
+	 * lock the proposed tablename against other would-be creators. The
+	 * insertion will roll back if we find problems below.
 	 */
-	rel = heap_openr(TableSpaceRelationName, RowExclusiveLock);
+	rel = heap_open(TableSpaceRelationId, RowExclusiveLock);
 
 	MemSet(nulls, ' ', Natts_pg_tablespace);
 
 	values[Anum_pg_tablespace_spcname - 1] =
 		DirectFunctionCall1(namein, CStringGetDatum(stmt->tablespacename));
 	values[Anum_pg_tablespace_spcowner - 1] =
-		Int32GetDatum(ownerid);
+		ObjectIdGetDatum(ownerId);
 	values[Anum_pg_tablespace_spclocation - 1] =
 		DirectFunctionCall1(textin, CStringGetDatum(location));
 	nulls[Anum_pg_tablespace_spcacl - 1] = 'n';
@@ -299,15 +308,18 @@ CreateTableSpace(CreateTableSpaceStmt *stmt)
 
 	heap_freetuple(tuple);
 
+	/* Record dependency on owner */
+	recordDependencyOnOwner(TableSpaceRelationId, tablespaceoid, ownerId);
+
 	/*
-	 * Attempt to coerce target directory to safe permissions.	If this
-	 * fails, it doesn't exist or has the wrong owner.
+	 * Attempt to coerce target directory to safe permissions.	If this fails,
+	 * it doesn't exist or has the wrong owner.
 	 */
 	if (chmod(location, 0700) != 0)
 		ereport(ERROR,
 				(errcode_for_file_access(),
-			  errmsg("could not set permissions on directory \"%s\": %m",
-					 location)));
+				 errmsg("could not set permissions on directory \"%s\": %m",
+						location)));
 
 	/*
 	 * Check the target directory is empty.
@@ -319,19 +331,19 @@ CreateTableSpace(CreateTableSpaceStmt *stmt)
 						location)));
 
 	/*
-	 * Create the PG_VERSION file in the target directory.	This has
-	 * several purposes: to make sure we can write in the directory, to
-	 * prevent someone from creating another tablespace pointing at the
-	 * same directory (the emptiness check above will fail), and to label
-	 * tablespace directories by PG version.
+	 * Create the PG_VERSION file in the target directory.	This has several
+	 * purposes: to make sure we can write in the directory, to prevent
+	 * someone from creating another tablespace pointing at the same directory
+	 * (the emptiness check above will fail), and to label tablespace
+	 * directories by PG version.
 	 */
 	set_short_version(location);
 
 	/*
 	 * All seems well, create the symlink
 	 */
-	linkloc = (char *) palloc(strlen(DataDir) + 11 + 10 + 1);
-	sprintf(linkloc, "%s/pg_tblspc/%u", DataDir, tablespaceoid);
+	linkloc = (char *) palloc(10 + 10 + 1);
+	sprintf(linkloc, "pg_tblspc/%u", tablespaceoid);
 
 	if (symlink(location, linkloc) < 0)
 		ereport(ERROR,
@@ -345,14 +357,14 @@ CreateTableSpace(CreateTableSpaceStmt *stmt)
 		XLogRecData rdata[2];
 
 		xlrec.ts_id = tablespaceoid;
-		rdata[0].buffer = InvalidBuffer;
 		rdata[0].data = (char *) &xlrec;
 		rdata[0].len = offsetof(xl_tblspc_create_rec, ts_path);
+		rdata[0].buffer = InvalidBuffer;
 		rdata[0].next = &(rdata[1]);
 
-		rdata[1].buffer = InvalidBuffer;
 		rdata[1].data = (char *) location;
 		rdata[1].len = strlen(location) + 1;
+		rdata[1].buffer = InvalidBuffer;
 		rdata[1].next = NULL;
 
 		(void) XLogInsert(RM_TBLSPC_ID, XLOG_TBLSPC_CREATE, rdata);
@@ -363,7 +375,6 @@ CreateTableSpace(CreateTableSpaceStmt *stmt)
 
 	/* We keep the lock on pg_tablespace until commit */
 	heap_close(rel, NoLock);
-
 #else							/* !HAVE_SYMLINK */
 	ereport(ERROR,
 			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
@@ -390,11 +401,15 @@ DropTableSpace(DropTableSpaceStmt *stmt)
 	/* don't call this in a transaction block */
 	PreventTransactionChain((void *) stmt, "DROP TABLESPACE");
 
+	/*
+	 * Acquire ExclusiveLock on pg_tablespace to ensure that no one else is
+	 * trying to do DROP TABLESPACE or TablespaceCreateDbspace concurrently.
+	 */
+	rel = heap_open(TableSpaceRelationId, ExclusiveLock);
+
 	/*
 	 * Find the target tuple
 	 */
-	rel = heap_openr(TableSpaceRelationName, RowExclusiveLock);
-
 	ScanKeyInit(&entry[0],
 				Anum_pg_tablespace_spcname,
 				BTEqualStrategyNumber, F_NAMEEQ,
@@ -410,9 +425,8 @@ DropTableSpace(DropTableSpaceStmt *stmt)
 
 	tablespaceoid = HeapTupleGetOid(tuple);
 
-	/* Must be superuser or owner */
-	if (GetUserId() != ((Form_pg_tablespace) GETSTRUCT(tuple))->spcowner &&
-		!superuser())
+	/* Must be tablespace owner */
+	if (!pg_tablespace_ownercheck(tablespaceoid, GetUserId()))
 		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_TABLESPACE,
 					   tablespacename);
 
@@ -423,18 +437,16 @@ DropTableSpace(DropTableSpaceStmt *stmt)
 					   tablespacename);
 
 	/*
-	 * Remove the pg_tablespace tuple (this will roll back if we fail
-	 * below)
+	 * Remove the pg_tablespace tuple (this will roll back if we fail below)
 	 */
 	simple_heap_delete(rel, &tuple->t_self);
 
 	heap_endscan(scandesc);
 
 	/*
-	 * Acquire TablespaceCreateLock to ensure that no TablespaceCreateDbspace
-	 * is running concurrently.
+	 * Remove dependency on owner.
 	 */
-	LWLockAcquire(TablespaceCreateLock, LW_EXCLUSIVE);
+	deleteSharedDependencyRecordsFor(TableSpaceRelationId, tablespaceoid);
 
 	/*
 	 * Try to remove the physical infrastructure
@@ -451,22 +463,16 @@ DropTableSpace(DropTableSpaceStmt *stmt)
 		XLogRecData rdata[1];
 
 		xlrec.ts_id = tablespaceoid;
-		rdata[0].buffer = InvalidBuffer;
 		rdata[0].data = (char *) &xlrec;
 		rdata[0].len = sizeof(xl_tblspc_drop_rec);
+		rdata[0].buffer = InvalidBuffer;
 		rdata[0].next = NULL;
 
 		(void) XLogInsert(RM_TBLSPC_ID, XLOG_TBLSPC_DROP, rdata);
 	}
 
-	/*
-	 * Allow TablespaceCreateDbspace again.
-	 */
-	LWLockRelease(TablespaceCreateLock);
-
 	/* We keep the lock on pg_tablespace until commit */
 	heap_close(rel, NoLock);
-
 #else							/* !HAVE_SYMLINK */
 	ereport(ERROR,
 			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
@@ -490,21 +496,21 @@ remove_tablespace_directories(Oid tablespaceoid, bool redo)
 	char	   *subfile;
 	struct stat st;
 
-	location = (char *) palloc(strlen(DataDir) + 11 + 10 + 1);
-	sprintf(location, "%s/pg_tblspc/%u", DataDir, tablespaceoid);
+	location = (char *) palloc(10 + 10 + 1);
+	sprintf(location, "pg_tblspc/%u", tablespaceoid);
 
 	/*
-	 * Check if the tablespace still contains any files.  We try to rmdir
-	 * each per-database directory we find in it.  rmdir failure implies
-	 * there are still files in that subdirectory, so give up.	(We do not
-	 * have to worry about undoing any already completed rmdirs, since the
-	 * next attempt to use the tablespace from that database will simply
-	 * recreate the subdirectory via TablespaceCreateDbspace.)
+	 * Check if the tablespace still contains any files.  We try to rmdir each
+	 * per-database directory we find in it.  rmdir failure implies there are
+	 * still files in that subdirectory, so give up.  (We do not have to worry
+	 * about undoing any already completed rmdirs, since the next attempt to
+	 * use the tablespace from that database will simply recreate the
+	 * subdirectory via TablespaceCreateDbspace.)
 	 *
-	 * Since we hold TablespaceCreateLock, no one else should be creating any
-	 * fresh subdirectories in parallel. It is possible that new files are
-	 * being created within subdirectories, though, so the rmdir call could
-	 * fail.  Worst consequence is a less friendly error message.
+	 * Since we hold exclusive lock, no one else should be creating any fresh
+	 * subdirectories in parallel.	It is possible that new files are being
+	 * created within subdirectories, though, so the rmdir call could fail.
+	 * Worst consequence is a less friendly error message.
 	 */
 	dirdesc = AllocateDir(location);
 	if (dirdesc == NULL)
@@ -514,23 +520,16 @@ remove_tablespace_directories(Oid tablespaceoid, bool redo)
 			pfree(location);
 			return true;
 		}
-		ereport(ERROR,
-				(errcode_for_file_access(),
-				 errmsg("could not open directory \"%s\": %m",
-						location)));
+		/* else let ReadDir report the error */
 	}
 
-	errno = 0;
-	while ((de = readdir(dirdesc)) != NULL)
+	while ((de = ReadDir(dirdesc, location)) != NULL)
 	{
 		/* Note we ignore PG_VERSION for the nonce */
 		if (strcmp(de->d_name, ".") == 0 ||
 			strcmp(de->d_name, "..") == 0 ||
 			strcmp(de->d_name, "PG_VERSION") == 0)
-		{
-			errno = 0;
 			continue;
-		}
 
 		subfile = palloc(strlen(location) + 1 + strlen(de->d_name) + 1);
 		sprintf(subfile, "%s/%s", location, de->d_name);
@@ -550,27 +549,13 @@ remove_tablespace_directories(Oid tablespaceoid, bool redo)
 							subfile)));
 
 		pfree(subfile);
-		errno = 0;
 	}
-#ifdef WIN32
 
-	/*
-	 * This fix is in mingw cvs (runtime/mingwex/dirent.c rev 1.4), but
-	 * not in released version
-	 */
-	if (GetLastError() == ERROR_NO_MORE_FILES)
-		errno = 0;
-#endif
-	if (errno)
-		ereport(ERROR,
-				(errcode_for_file_access(),
-				 errmsg("could not read directory \"%s\": %m",
-						location)));
 	FreeDir(dirdesc);
 
 	/*
-	 * Okay, try to unlink PG_VERSION (we allow it to not be there, even
-	 * in non-REDO case, for robustness).
+	 * Okay, try to unlink PG_VERSION (we allow it to not be there, even in
+	 * non-REDO case, for robustness).
 	 */
 	subfile = palloc(strlen(location) + 11 + 1);
 	sprintf(subfile, "%s/PG_VERSION", location);
@@ -588,9 +573,9 @@ remove_tablespace_directories(Oid tablespaceoid, bool redo)
 
 	/*
 	 * Okay, try to remove the symlink.  We must however deal with the
-	 * possibility that it's a directory instead of a symlink --- this
-	 * could happen during WAL replay (see TablespaceCreateDbspace), and
-	 * it is also the normal case on Windows.
+	 * possibility that it's a directory instead of a symlink --- this could
+	 * happen during WAL replay (see TablespaceCreateDbspace), and it is also
+	 * the normal case on Windows.
 	 */
 	if (lstat(location, &st) == 0 && S_ISDIR(st.st_mode))
 	{
@@ -680,38 +665,16 @@ directory_is_empty(const char *path)
 	struct dirent *de;
 
 	dirdesc = AllocateDir(path);
-	if (dirdesc == NULL)
-		ereport(ERROR,
-				(errcode_for_file_access(),
-				 errmsg("could not open directory \"%s\": %m",
-						path)));
 
-	errno = 0;
-	while ((de = readdir(dirdesc)) != NULL)
+	while ((de = ReadDir(dirdesc, path)) != NULL)
 	{
 		if (strcmp(de->d_name, ".") == 0 ||
 			strcmp(de->d_name, "..") == 0)
-		{
-			errno = 0;
 			continue;
-		}
 		FreeDir(dirdesc);
 		return false;
 	}
-#ifdef WIN32
 
-	/*
-	 * This fix is in mingw cvs (runtime/mingwex/dirent.c rev 1.4), but
-	 * not in released version
-	 */
-	if (GetLastError() == ERROR_NO_MORE_FILES)
-		errno = 0;
-#endif
-	if (errno)
-		ereport(ERROR,
-				(errcode_for_file_access(),
-				 errmsg("could not read directory \"%s\": %m",
-						path)));
 	FreeDir(dirdesc);
 	return true;
 }
@@ -730,7 +693,7 @@ RenameTableSpace(const char *oldname, const char *newname)
 	Form_pg_tablespace newform;
 
 	/* Search pg_tablespace */
-	rel = heap_openr(TableSpaceRelationName, RowExclusiveLock);
+	rel = heap_open(TableSpaceRelationId, RowExclusiveLock);
 
 	ScanKeyInit(&entry[0],
 				Anum_pg_tablespace_spcname,
@@ -749,8 +712,8 @@ RenameTableSpace(const char *oldname, const char *newname)
 
 	heap_endscan(scan);
 
-	/* Must be owner or superuser */
-	if (newform->spcowner != GetUserId() && !superuser())
+	/* Must be owner */
+	if (!pg_tablespace_ownercheck(HeapTupleGetOid(newtuple), GetUserId()))
 		aclcheck_error(ACLCHECK_NO_PRIV, ACL_KIND_TABLESPACE, oldname);
 
 	/* Validate new name */
@@ -758,7 +721,7 @@ RenameTableSpace(const char *oldname, const char *newname)
 		ereport(ERROR,
 				(errcode(ERRCODE_RESERVED_NAME),
 				 errmsg("unacceptable tablespace name \"%s\"", newname),
-				 errdetail("The prefix \"pg_\" is reserved for system tablespaces.")));
+		errdetail("The prefix \"pg_\" is reserved for system tablespaces.")));
 
 	/* Make sure the new name doesn't exist */
 	ScanKeyInit(&entry[0],
@@ -788,7 +751,7 @@ RenameTableSpace(const char *oldname, const char *newname)
  * Change tablespace owner
  */
 void
-AlterTableSpaceOwner(const char *name, AclId newOwnerSysId)
+AlterTableSpaceOwner(const char *name, Oid newOwnerId)
 {
 	Relation	rel;
 	ScanKeyData entry[1];
@@ -797,7 +760,7 @@ AlterTableSpaceOwner(const char *name, AclId newOwnerSysId)
 	HeapTuple	tup;
 
 	/* Search pg_tablespace */
-	rel = heap_openr(TableSpaceRelationName, RowExclusiveLock);
+	rel = heap_open(TableSpaceRelationId, RowExclusiveLock);
 
 	ScanKeyInit(&entry[0],
 				Anum_pg_tablespace_spcname,
@@ -816,7 +779,7 @@ AlterTableSpaceOwner(const char *name, AclId newOwnerSysId)
 	 * If the new owner is the same as the existing owner, consider the
 	 * command to have succeeded.  This is for dump restoration purposes.
 	 */
-	if (spcForm->spcowner != newOwnerSysId)
+	if (spcForm->spcowner != newOwnerId)
 	{
 		Datum		repl_val[Natts_pg_tablespace];
 		char		repl_null[Natts_pg_tablespace];
@@ -826,17 +789,29 @@ AlterTableSpaceOwner(const char *name, AclId newOwnerSysId)
 		bool		isNull;
 		HeapTuple	newtuple;
 
-		/* Otherwise, must be superuser to change object ownership */
-		if (!superuser())
-			ereport(ERROR,
-					(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
-					 errmsg("must be superuser to change owner")));
+		/* Otherwise, must be owner of the existing object */
+		if (!pg_tablespace_ownercheck(HeapTupleGetOid(tup), GetUserId()))
+			aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_TABLESPACE,
+						   name);
+
+		/* Must be able to become new owner */
+		check_is_member_of_role(GetUserId(), newOwnerId);
+
+		/*
+		 * Normally we would also check for create permissions here, but there
+		 * are none for tablespaces so we follow what rename tablespace does
+		 * and omit the create permissions check.
+		 *
+		 * NOTE: Only superusers may create tablespaces to begin with and so
+		 * initially only a superuser would be able to change its ownership
+		 * anyway.
+		 */
 
 		memset(repl_null, ' ', sizeof(repl_null));
 		memset(repl_repl, ' ', sizeof(repl_repl));
 
 		repl_repl[Anum_pg_tablespace_spcowner - 1] = 'r';
-		repl_val[Anum_pg_tablespace_spcowner - 1] = Int32GetDatum(newOwnerSysId);
+		repl_val[Anum_pg_tablespace_spcowner - 1] = ObjectIdGetDatum(newOwnerId);
 
 		/*
 		 * Determine the modified ACL for the new owner.  This is only
@@ -849,17 +824,21 @@ AlterTableSpaceOwner(const char *name, AclId newOwnerSysId)
 		if (!isNull)
 		{
 			newAcl = aclnewowner(DatumGetAclP(aclDatum),
-								 spcForm->spcowner, newOwnerSysId);
+								 spcForm->spcowner, newOwnerId);
 			repl_repl[Anum_pg_tablespace_spcacl - 1] = 'r';
 			repl_val[Anum_pg_tablespace_spcacl - 1] = PointerGetDatum(newAcl);
 		}
 
-		newtuple = heap_modifytuple(tup, rel, repl_val, repl_null, repl_repl);
+		newtuple = heap_modifytuple(tup, RelationGetDescr(rel), repl_val, repl_null, repl_repl);
 
 		simple_heap_update(rel, &newtuple->t_self, newtuple);
 		CatalogUpdateIndexes(rel, newtuple);
 
 		heap_freetuple(newtuple);
+
+		/* Update owner dependency reference */
+		changeDependencyOnOwner(TableSpaceRelationId, HeapTupleGetOid(tup),
+								newOwnerId);
 	}
 
 	heap_endscan(scandesc);
@@ -877,7 +856,7 @@ assign_default_tablespace(const char *newval, bool doit, GucSource source)
 {
 	/*
 	 * If we aren't inside a transaction, we cannot do database access so
-	 * cannot verify the name.  Must accept the value on faith.
+	 * cannot verify the name.	Must accept the value on faith.
 	 */
 	if (IsTransactionState())
 	{
@@ -912,15 +891,16 @@ GetDefaultTablespace(void)
 	/* Fast path for default_tablespace == "" */
 	if (default_tablespace == NULL || default_tablespace[0] == '\0')
 		return InvalidOid;
+
 	/*
 	 * It is tempting to cache this lookup for more speed, but then we would
-	 * fail to detect the case where the tablespace was dropped since the
-	 * GUC variable was set.  Note also that we don't complain if the value
-	 * fails to refer to an existing tablespace; we just silently return
-	 * InvalidOid, causing the new object to be created in the database's
-	 * tablespace.
+	 * fail to detect the case where the tablespace was dropped since the GUC
+	 * variable was set.  Note also that we don't complain if the value fails
+	 * to refer to an existing tablespace; we just silently return InvalidOid,
+	 * causing the new object to be created in the database's tablespace.
 	 */
 	result = get_tablespace_oid(default_tablespace);
+
 	/*
 	 * Allow explicit specification of database's default tablespace in
 	 * default_tablespace without triggering permissions checks.
@@ -946,7 +926,7 @@ get_tablespace_oid(const char *tablespacename)
 	ScanKeyData entry[1];
 
 	/* Search pg_tablespace */
-	rel = heap_openr(TableSpaceRelationName, AccessShareLock);
+	rel = heap_open(TableSpaceRelationId, AccessShareLock);
 
 	ScanKeyInit(&entry[0],
 				Anum_pg_tablespace_spcname,
@@ -981,7 +961,7 @@ get_tablespace_name(Oid spc_oid)
 	ScanKeyData entry[1];
 
 	/* Search pg_tablespace */
-	rel = heap_openr(TableSpaceRelationName, AccessShareLock);
+	rel = heap_open(TableSpaceRelationId, AccessShareLock);
 
 	ScanKeyInit(&entry[0],
 				ObjectIdAttributeNumber,
@@ -1018,29 +998,29 @@ tblspc_redo(XLogRecPtr lsn, XLogRecord *record)
 		char	   *linkloc;
 
 		/*
-		 * Attempt to coerce target directory to safe permissions.	If
-		 * this fails, it doesn't exist or has the wrong owner.
+		 * Attempt to coerce target directory to safe permissions.	If this
+		 * fails, it doesn't exist or has the wrong owner.
 		 */
 		if (chmod(location, 0700) != 0)
 			ereport(ERROR,
 					(errcode_for_file_access(),
-			  errmsg("could not set permissions on directory \"%s\": %m",
-					 location)));
+				  errmsg("could not set permissions on directory \"%s\": %m",
+						 location)));
 
 		/* Create or re-create the PG_VERSION file in the target directory */
 		set_short_version(location);
 
 		/* Create the symlink if not already present */
-		linkloc = (char *) palloc(strlen(DataDir) + 11 + 10 + 1);
-		sprintf(linkloc, "%s/pg_tblspc/%u", DataDir, xlrec->ts_id);
+		linkloc = (char *) palloc(10 + 10 + 1);
+		sprintf(linkloc, "pg_tblspc/%u", xlrec->ts_id);
 
 		if (symlink(location, linkloc) < 0)
 		{
 			if (errno != EEXIST)
 				ereport(ERROR,
 						(errcode_for_file_access(),
-					  errmsg("could not create symbolic link \"%s\": %m",
-							 linkloc)));
+						 errmsg("could not create symbolic link \"%s\": %m",
+								linkloc)));
 		}
 
 		pfree(linkloc);
@@ -1059,12 +1039,6 @@ tblspc_redo(XLogRecPtr lsn, XLogRecord *record)
 		elog(PANIC, "tblspc_redo: unknown op code %u", info);
 }
 
-void
-tblspc_undo(XLogRecPtr lsn, XLogRecord *record)
-{
-	elog(PANIC, "tblspc_undo: unimplemented");
-}
-
 void
 tblspc_desc(char *buf, uint8 xl_info, char *rec)
 {
