@@ -6,12 +6,12 @@
  *	  message integrity and endpoint authentication.
  *
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/libpq/be-secure.c,v 1.43.2.3 2006/05/12 22:44:58 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/libpq/be-secure.c,v 1.59 2005/10/15 02:49:17 momjian Exp $
  *
  *	  Since the server static private key ($DataDir/server.key)
  *	  will normally be stored unencrypted so that the database
@@ -100,6 +100,11 @@
 
 
 #ifdef USE_SSL
+
+#define ROOT_CERT_FILE			"root.crt"
+#define SERVER_CERT_FILE		"server.crt"
+#define SERVER_PRIVATE_KEY_FILE "server.key"
+
 static DH  *load_dh_file(int keylength);
 static DH  *load_dh_buffer(const char *, size_t);
 static DH  *tmp_dh_cb(SSL *s, int is_export, int keylength);
@@ -118,7 +123,6 @@ static const char *SSLerrmessage(void);
  *	(total in both directions) before we require renegotiation.
  */
 #define RENEGOTIATION_LIMIT (512 * 1024 * 1024)
-#define CA_PATH NULL
 
 static SSL_CTX *SSL_context = NULL;
 #endif
@@ -258,15 +262,23 @@ secure_read(Port *port, void *ptr, size_t len)
 #ifdef USE_SSL
 	if (port->ssl)
 	{
+		int			err;
+
 rloop:
 		n = SSL_read(port->ssl, ptr, len);
-		switch (SSL_get_error(port->ssl, n))
+		err = SSL_get_error(port->ssl, n);
+		switch (err)
 		{
 			case SSL_ERROR_NONE:
 				port->count += n;
 				break;
 			case SSL_ERROR_WANT_READ:
 			case SSL_ERROR_WANT_WRITE:
+#ifdef WIN32
+				pgwin32_waitforsinglesocket(SSL_get_fd(port->ssl),
+											(err == SSL_ERROR_WANT_READ) ?
+								   FD_READ | FD_CLOSE : FD_WRITE | FD_CLOSE);
+#endif
 				goto rloop;
 			case SSL_ERROR_SYSCALL:
 				if (n == -1)
@@ -294,8 +306,8 @@ rloop:
 			default:
 				ereport(COMMERROR,
 						(errcode(ERRCODE_PROTOCOL_VIOLATION),
-						 errmsg("unrecognized SSL error code %d",
-							SSL_get_error(port->ssl, n))));
+						 errmsg("unrecognized SSL error code: %d",
+								err)));
 				n = -1;
 				break;
 		}
@@ -324,6 +336,8 @@ secure_write(Port *port, void *ptr, size_t len)
 #ifdef USE_SSL
 	if (port->ssl)
 	{
+		int			err;
+
 		if (port->count > RENEGOTIATION_LIMIT)
 		{
 			SSL_set_session_id_context(port->ssl, (void *) &SSL_context,
@@ -339,7 +353,7 @@ secure_write(Port *port, void *ptr, size_t len)
 			if (port->ssl->state != SSL_ST_OK)
 				ereport(COMMERROR,
 						(errcode(ERRCODE_PROTOCOL_VIOLATION),
-					errmsg("SSL failed to send renegotiation request")));
+						 errmsg("SSL failed to send renegotiation request")));
 			port->ssl->state |= SSL_ST_ACCEPT;
 			SSL_do_handshake(port->ssl);
 			if (port->ssl->state != SSL_ST_OK)
@@ -351,13 +365,19 @@ secure_write(Port *port, void *ptr, size_t len)
 
 wloop:
 		n = SSL_write(port->ssl, ptr, len);
-		switch (SSL_get_error(port->ssl, n))
+		err = SSL_get_error(port->ssl, n);
+		switch (err)
 		{
 			case SSL_ERROR_NONE:
 				port->count += n;
 				break;
 			case SSL_ERROR_WANT_READ:
 			case SSL_ERROR_WANT_WRITE:
+#ifdef WIN32
+				pgwin32_waitforsinglesocket(SSL_get_fd(port->ssl),
+											(err == SSL_ERROR_WANT_READ) ?
+								   FD_READ | FD_CLOSE : FD_WRITE | FD_CLOSE);
+#endif
 				goto wloop;
 			case SSL_ERROR_SYSCALL:
 				if (n == -1)
@@ -385,8 +405,8 @@ wloop:
 			default:
 				ereport(COMMERROR,
 						(errcode(ERRCODE_PROTOCOL_VIOLATION),
-						 errmsg("unrecognized SSL error code %d",
-							SSL_get_error(port->ssl, n))));
+						 errmsg("unrecognized SSL error code: %d",
+								err)));
 				n = -1;
 				break;
 		}
@@ -419,12 +439,12 @@ wloop:
 
 static bool my_bio_initialized = false;
 static BIO_METHOD my_bio_methods;
-static int (*std_sock_read) (BIO *h, char *buf, int size);
+static int	(*std_sock_read) (BIO *h, char *buf, int size);
 
 static int
 my_sock_read(BIO *h, char *buf, int size)
 {
-	int		res;
+	int			res;
 
 	prepare_for_client_read();
 
@@ -452,21 +472,21 @@ my_BIO_s_socket(void)
 static int
 my_SSL_set_fd(SSL *s, int fd)
 {
-	int ret=0;
-	BIO *bio=NULL;
+	int			ret = 0;
+	BIO		   *bio = NULL;
 
-	bio=BIO_new(my_BIO_s_socket());
+	bio = BIO_new(my_BIO_s_socket());
 
 	if (bio == NULL)
 	{
-		SSLerr(SSL_F_SSL_SET_FD,ERR_R_BUF_LIB);
+		SSLerr(SSL_F_SSL_SET_FD, ERR_R_BUF_LIB);
 		goto err;
 	}
-	BIO_set_fd(bio,fd,BIO_NOCLOSE);
-	SSL_set_bio(s,bio,bio);
-	ret=1;
+	BIO_set_fd(bio, fd, BIO_NOCLOSE);
+	SSL_set_bio(s, bio, bio);
+	ret = 1;
 err:
-	return(ret);
+	return (ret);
 }
 
 /*
@@ -480,12 +500,12 @@ static DH  *
 load_dh_file(int keylength)
 {
 	FILE	   *fp;
-	char		fnbuf[2048];
+	char		fnbuf[MAXPGPATH];
 	DH		   *dh = NULL;
 	int			codes;
 
 	/* attempt to open file.  It's not an error if it doesn't exist. */
-	snprintf(fnbuf, sizeof fnbuf, "%s/dh%d.pem", DataDir, keylength);
+	snprintf(fnbuf, sizeof(fnbuf), "dh%d.pem", keylength);
 	if ((fp = fopen(fnbuf, "r")) == NULL)
 		return NULL;
 
@@ -505,7 +525,7 @@ load_dh_file(int keylength)
 	/* make sure the DH parameters are usable */
 	if (dh != NULL)
 	{
-		if (DH_check(dh, &codes) == 0)
+		if (DH_check(dh, &codes))
 		{
 			elog(LOG, "DH_check error (%s): %s", fnbuf, SSLerrmessage());
 			return NULL;
@@ -519,7 +539,7 @@ load_dh_file(int keylength)
 			(codes & DH_CHECK_P_NOT_SAFE_PRIME))
 		{
 			elog(LOG,
-			   "DH error (%s): neither suitable generator or safe prime",
+				 "DH error (%s): neither suitable generator or safe prime",
 				 fnbuf);
 			return NULL;
 		}
@@ -695,7 +715,6 @@ info_cb(const SSL *ssl, int type, int args)
 static int
 initialize_SSL(void)
 {
-	char		fnbuf[MAXPGPATH];
 	struct stat buf;
 
 	if (!SSL_context)
@@ -711,31 +730,44 @@ initialize_SSL(void)
 		/*
 		 * Load and verify certificate and private key
 		 */
-		snprintf(fnbuf, sizeof(fnbuf), "%s/server.crt", DataDir);
-		if (!SSL_CTX_use_certificate_file(SSL_context, fnbuf, SSL_FILETYPE_PEM))
+		if (!SSL_CTX_use_certificate_file(SSL_context,
+										  SERVER_CERT_FILE,
+										  SSL_FILETYPE_PEM))
 			ereport(FATAL,
 					(errcode(ERRCODE_CONFIG_FILE_ERROR),
-			  errmsg("could not load server certificate file \"%s\": %s",
-					 fnbuf, SSLerrmessage())));
+				  errmsg("could not load server certificate file \"%s\": %s",
+						 SERVER_CERT_FILE, SSLerrmessage())));
 
-		snprintf(fnbuf, sizeof(fnbuf), "%s/server.key", DataDir);
-		if (stat(fnbuf, &buf) == -1)
+		if (stat(SERVER_PRIVATE_KEY_FILE, &buf) == -1)
 			ereport(FATAL,
 					(errcode_for_file_access(),
-				   errmsg("could not access private key file \"%s\": %m",
-						  fnbuf)));
+					 errmsg("could not access private key file \"%s\": %m",
+							SERVER_PRIVATE_KEY_FILE)));
+
+		/*
+		 * Require no public access to key file.
+		 *
+		 * XXX temporarily suppress check when on Windows, because there may not
+		 * be proper support for Unix-y file permissions.  Need to think of a
+		 * reasonable check to apply on Windows.  (See also the data directory
+		 * permission check in postmaster.c)
+		 */
+#if !defined(WIN32) && !defined(__CYGWIN__)
 		if (!S_ISREG(buf.st_mode) || (buf.st_mode & (S_IRWXG | S_IRWXO)) ||
-			buf.st_uid != getuid())
+			buf.st_uid != geteuid())
 			ereport(FATAL,
 					(errcode(ERRCODE_CONFIG_FILE_ERROR),
-				  errmsg("unsafe permissions on private key file \"%s\"",
-						 fnbuf),
+					 errmsg("unsafe permissions on private key file \"%s\"",
+							SERVER_PRIVATE_KEY_FILE),
 					 errdetail("File must be owned by the database user and must have no permissions for \"group\" or \"other\".")));
+#endif
 
-		if (!SSL_CTX_use_PrivateKey_file(SSL_context, fnbuf, SSL_FILETYPE_PEM))
+		if (!SSL_CTX_use_PrivateKey_file(SSL_context,
+										 SERVER_PRIVATE_KEY_FILE,
+										 SSL_FILETYPE_PEM))
 			ereport(FATAL,
 					(errmsg("could not load private key file \"%s\": %s",
-							fnbuf, SSLerrmessage())));
+							SERVER_PRIVATE_KEY_FILE, SSLerrmessage())));
 
 		if (!SSL_CTX_check_private_key(SSL_context))
 			ereport(FATAL,
@@ -751,20 +783,25 @@ initialize_SSL(void)
 	if (SSL_CTX_set_cipher_list(SSL_context, "ALL:!ADH:!LOW:!EXP:!MD5:@STRENGTH") != 1)
 		elog(FATAL, "could not set the cipher list (no valid ciphers available)");
 
-	/* accept client certificates, but don't require them. */
-	snprintf(fnbuf, sizeof(fnbuf), "%s/root.crt", DataDir);
-	if (!SSL_CTX_load_verify_locations(SSL_context, fnbuf, CA_PATH))
+	/*
+	 * Require and check client certificates only if we have a root.crt file.
+	 */
+	if (!SSL_CTX_load_verify_locations(SSL_context, ROOT_CERT_FILE, NULL))
 	{
 		/* Not fatal - we do not require client certificates */
 		ereport(LOG,
 				(errmsg("could not load root certificate file \"%s\": %s",
-						fnbuf, SSLerrmessage()),
+						ROOT_CERT_FILE, SSLerrmessage()),
 				 errdetail("Will not verify client certificates.")));
-		return 0;
 	}
-	SSL_CTX_set_verify(SSL_context,
-					   SSL_VERIFY_PEER | SSL_VERIFY_CLIENT_ONCE,
-					   verify_cb);
+	else
+	{
+		SSL_CTX_set_verify(SSL_context,
+						   (SSL_VERIFY_PEER |
+							SSL_VERIFY_FAIL_IF_NO_PEER_CERT |
+							SSL_VERIFY_CLIENT_ONCE),
+						   verify_cb);
+	}
 
 	return 0;
 }
@@ -788,12 +825,13 @@ destroy_SSL(void)
 static int
 open_server_SSL(Port *port)
 {
+	int			r;
+	int			err;
+
 	Assert(!port->ssl);
 	Assert(!port->peer);
 
-	if (!(port->ssl = SSL_new(SSL_context)) ||
-		!my_SSL_set_fd(port->ssl, port->sock) ||
-		SSL_accept(port->ssl) <= 0)
+	if (!(port->ssl = SSL_new(SSL_context)))
 	{
 		ereport(COMMERROR,
 				(errcode(ERRCODE_PROTOCOL_VIOLATION),
@@ -802,6 +840,63 @@ open_server_SSL(Port *port)
 		close_SSL(port);
 		return -1;
 	}
+	if (!my_SSL_set_fd(port->ssl, port->sock))
+	{
+		ereport(COMMERROR,
+				(errcode(ERRCODE_PROTOCOL_VIOLATION),
+				 errmsg("could not set SSL socket: %s",
+						SSLerrmessage())));
+		close_SSL(port);
+		return -1;
+	}
+
+aloop:
+	r = SSL_accept(port->ssl);
+	if (r <= 0)
+	{
+		err = SSL_get_error(port->ssl, r);
+		switch (err)
+		{
+			case SSL_ERROR_WANT_READ:
+			case SSL_ERROR_WANT_WRITE:
+#ifdef WIN32
+				pgwin32_waitforsinglesocket(SSL_get_fd(port->ssl),
+											(err == SSL_ERROR_WANT_READ) ?
+					   FD_READ | FD_CLOSE | FD_ACCEPT : FD_WRITE | FD_CLOSE);
+#endif
+				goto aloop;
+			case SSL_ERROR_SYSCALL:
+				if (r < 0)
+					ereport(COMMERROR,
+							(errcode_for_socket_access(),
+							 errmsg("could not accept SSL connection: %m")));
+				else
+					ereport(COMMERROR,
+							(errcode(ERRCODE_PROTOCOL_VIOLATION),
+					errmsg("could not accept SSL connection: EOF detected")));
+				break;
+			case SSL_ERROR_SSL:
+				ereport(COMMERROR,
+						(errcode(ERRCODE_PROTOCOL_VIOLATION),
+						 errmsg("could not accept SSL connection: %s",
+								SSLerrmessage())));
+				break;
+			case SSL_ERROR_ZERO_RETURN:
+				ereport(COMMERROR,
+						(errcode(ERRCODE_PROTOCOL_VIOLATION),
+				   errmsg("could not accept SSL connection: EOF detected")));
+				break;
+			default:
+				ereport(COMMERROR,
+						(errcode(ERRCODE_PROTOCOL_VIOLATION),
+						 errmsg("unrecognized SSL error code: %d",
+								err)));
+				break;
+		}
+		close_SSL(port);
+		return -1;
+	}
+
 	port->count = 0;
 
 	/* get client certificate, if available. */
@@ -817,7 +912,7 @@ open_server_SSL(Port *port)
 						  port->peer_dn, sizeof(port->peer_dn));
 		port->peer_dn[sizeof(port->peer_dn) - 1] = '\0';
 		X509_NAME_get_text_by_NID(X509_get_subject_name(port->peer),
-				   NID_commonName, port->peer_cn, sizeof(port->peer_cn));
+					   NID_commonName, port->peer_cn, sizeof(port->peer_cn));
 		port->peer_cn[sizeof(port->peer_cn) - 1] = '\0';
 	}
 	ereport(DEBUG2,
