@@ -4,7 +4,7 @@
  * (currently mule internal code (mic) is used)
  * Tatsuo Ishii
  *
- * $Header: /cvsroot/pgsql/src/backend/utils/mb/mbutils.c,v 1.44 2003/09/25 06:58:05 petere Exp $
+ * $PostgreSQL: pgsql/src/backend/utils/mb/mbutils.c,v 1.52 2005/10/15 02:49:33 momjian Exp $
  */
 #include "postgres.h"
 
@@ -43,9 +43,9 @@ static int	pending_client_encoding = PG_SQL_ASCII;
 
 
 /* Internal functions */
-static unsigned char *perform_default_encoding_conversion(unsigned char *src,
+static char *perform_default_encoding_conversion(const char *src,
 									int len, bool is_client_to_server);
-static int	cliplen(const unsigned char *str, int len, int limit);
+static int	cliplen(const char *str, int len, int limit);
 
 
 /*
@@ -107,12 +107,11 @@ SetClientEncoding(int encoding, bool doit)
 	}
 
 	/*
-	 * If we're not inside a transaction then we can't do catalog lookups,
-	 * so fail.  After backend startup, this could only happen if we are
+	 * If we're not inside a transaction then we can't do catalog lookups, so
+	 * fail.  After backend startup, this could only happen if we are
 	 * re-reading postgresql.conf due to SIGHUP --- so basically this just
 	 * constrains the ability to change client_encoding on the fly from
-	 * postgresql.conf.  Which would probably be a stupid thing to do
-	 * anyway.
+	 * postgresql.conf.  Which would probably be a stupid thing to do anyway.
 	 */
 	if (!IsTransactionState())
 		return -1;
@@ -136,8 +135,8 @@ SetClientEncoding(int encoding, bool doit)
 		return 0;
 
 	/*
-	 * load the fmgr info into TopMemoryContext so that it survives
-	 * outside transaction.
+	 * load the fmgr info into TopMemoryContext so that it survives outside
+	 * transaction.
 	 */
 	oldcontext = MemoryContextSwitchTo(TopMemoryContext);
 	to_server = palloc(sizeof(FmgrInfo));
@@ -180,8 +179,8 @@ InitializeClientEncoding(void)
 	if (SetClientEncoding(pending_client_encoding, true) < 0)
 	{
 		/*
-		 * Oops, the requested conversion is not available. We couldn't
-		 * fail before, but we can now.
+		 * Oops, the requested conversion is not available. We couldn't fail
+		 * before, but we can now.
 		 */
 		ereport(FATAL,
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
@@ -222,7 +221,7 @@ pg_get_client_encoding_name(void)
  *
  * XXX We assume that storage for converted result is 4-to-1 growth in
  * the worst case. The rate for currently supported encoding pares are within 3
- * (SJIS JIS X0201 half width kanna -> UTF-8 is the worst case).
+ * (SJIS JIS X0201 half width kanna -> UTF8 is the worst case).
  * So "4" should be enough for the moment.
  */
 unsigned char *
@@ -249,15 +248,15 @@ pg_do_encoding_conversion(unsigned char *src, int len,
 	{
 		ereport(LOG,
 				(errcode(ERRCODE_UNDEFINED_FUNCTION),
-			errmsg("default conversion function for encoding \"%s\" to \"%s\" does not exist",
-				   pg_encoding_to_char(src_encoding),
-				   pg_encoding_to_char(dest_encoding))));
+				 errmsg("default conversion function for encoding \"%s\" to \"%s\" does not exist",
+						pg_encoding_to_char(src_encoding),
+						pg_encoding_to_char(dest_encoding))));
 		return src;
 	}
 
 	/*
-	 * XXX we should avoid throwing errors in OidFunctionCall. Otherwise
-	 * we are going into infinite loop!  So we have to make sure that the
+	 * XXX we should avoid throwing errors in OidFunctionCall. Otherwise we
+	 * are going into infinite loop!  So we have to make sure that the
 	 * function exists before calling OidFunctionCall.
 	 */
 	if (!SearchSysCacheExists(PROCOID,
@@ -290,11 +289,11 @@ pg_convert(PG_FUNCTION_ARGS)
 	Datum		string = PG_GETARG_DATUM(0);
 	Datum		dest_encoding_name = PG_GETARG_DATUM(1);
 	Datum		src_encoding_name = DirectFunctionCall1(
-						namein, CStringGetDatum(DatabaseEncoding->name));
+							namein, CStringGetDatum(DatabaseEncoding->name));
 	Datum		result;
 
 	result = DirectFunctionCall3(
-			 pg_convert2, string, src_encoding_name, dest_encoding_name);
+				 pg_convert2, string, src_encoding_name, dest_encoding_name);
 
 	/* free memory allocated by namein */
 	pfree((void *) src_encoding_name);
@@ -303,7 +302,7 @@ pg_convert(PG_FUNCTION_ARGS)
 }
 
 /*
- * Convert string using encoding_nanme.
+ * Convert string using encoding_name.
  *
  * TEXT convert2(TEXT string, NAME src_encoding_name, NAME dest_encoding_name)
  */
@@ -343,10 +342,9 @@ pg_convert2(PG_FUNCTION_ARGS)
 
 	/*
 	 * build text data type structure. we cannot use textin() here, since
-	 * textin assumes that input string encoding is same as database
-	 * encoding.
+	 * textin assumes that input string encoding is same as database encoding.
 	 */
-	len = strlen(result) + VARHDRSZ;
+	len = strlen((char *) result) + VARHDRSZ;
 	retval = palloc(len);
 	VARATT_SIZEP(retval) = len;
 	memcpy(VARDATA(retval), result, len - VARHDRSZ);
@@ -364,14 +362,14 @@ pg_convert2(PG_FUNCTION_ARGS)
 /*
  * convert client encoding to server encoding.
  */
-unsigned char *
-pg_client_to_server(unsigned char *s, int len)
+char *
+pg_client_to_server(const char *s, int len)
 {
 	Assert(DatabaseEncoding);
 	Assert(ClientEncoding);
 
 	if (ClientEncoding->encoding == DatabaseEncoding->encoding)
-		return s;
+		return (char *) s;
 
 	return perform_default_encoding_conversion(s, len, true);
 }
@@ -379,14 +377,14 @@ pg_client_to_server(unsigned char *s, int len)
 /*
  * convert server encoding to client encoding.
  */
-unsigned char *
-pg_server_to_client(unsigned char *s, int len)
+char *
+pg_server_to_client(const char *s, int len)
 {
 	Assert(DatabaseEncoding);
 	Assert(ClientEncoding);
 
 	if (ClientEncoding->encoding == DatabaseEncoding->encoding)
-		return s;
+		return (char *) s;
 
 	return perform_default_encoding_conversion(s, len, false);
 }
@@ -398,16 +396,16 @@ pg_server_to_client(unsigned char *s, int len)
  *	before calling this function. Otherwise no conversion is
  *	performed.
 */
-static unsigned char *
-perform_default_encoding_conversion(unsigned char *src, int len, bool is_client_to_server)
+static char *
+perform_default_encoding_conversion(const char *src, int len, bool is_client_to_server)
 {
-	unsigned char *result;
+	char	   *result;
 	int			src_encoding,
 				dest_encoding;
 	FmgrInfo   *flinfo;
 
 	if (len <= 0)
-		return src;
+		return (char *) src;
 
 	if (is_client_to_server)
 	{
@@ -423,13 +421,13 @@ perform_default_encoding_conversion(unsigned char *src, int len, bool is_client_
 	}
 
 	if (flinfo == NULL)
-		return src;
+		return (char *) src;
 
 	if (src_encoding == dest_encoding)
-		return src;
+		return (char *) src;
 
 	if (src_encoding == PG_SQL_ASCII || dest_encoding == PG_SQL_ASCII)
-		return src;
+		return (char *) src;
 
 	result = palloc(len * 4 + 1);
 
@@ -444,34 +442,41 @@ perform_default_encoding_conversion(unsigned char *src, int len, bool is_client_
 
 /* convert a multibyte string to a wchar */
 int
-pg_mb2wchar(const unsigned char *from, pg_wchar *to)
+pg_mb2wchar(const char *from, pg_wchar *to)
 {
-	return (*pg_wchar_table[DatabaseEncoding->encoding].mb2wchar_with_len) (from, to, strlen(from));
+	return (*pg_wchar_table[DatabaseEncoding->encoding].mb2wchar_with_len) ((const unsigned char *) from, to, strlen(from));
 }
 
 /* convert a multibyte string to a wchar with a limited length */
 int
-pg_mb2wchar_with_len(const unsigned char *from, pg_wchar *to, int len)
+pg_mb2wchar_with_len(const char *from, pg_wchar *to, int len)
 {
-	return (*pg_wchar_table[DatabaseEncoding->encoding].mb2wchar_with_len) (from, to, len);
+	return (*pg_wchar_table[DatabaseEncoding->encoding].mb2wchar_with_len) ((const unsigned char *) from, to, len);
 }
 
 /* returns the byte length of a multibyte word */
 int
-pg_mblen(const unsigned char *mbstr)
+pg_mblen(const char *mbstr)
 {
-	return ((*pg_wchar_table[DatabaseEncoding->encoding].mblen) (mbstr));
+	return ((*pg_wchar_table[DatabaseEncoding->encoding].mblen) ((const unsigned char *) mbstr));
 }
 
-/* returns the length (counted as a wchar) of a multibyte string */
+/* returns the display length of a multibyte word */
 int
-pg_mbstrlen(const unsigned char *mbstr)
+pg_dsplen(const char *mbstr)
+{
+	return ((*pg_wchar_table[DatabaseEncoding->encoding].dsplen) ((const unsigned char *) mbstr));
+}
+
+/* returns the length (counted in wchars) of a multibyte string */
+int
+pg_mbstrlen(const char *mbstr)
 {
 	int			len = 0;
 
 	/* optimization for single byte encoding */
 	if (pg_database_encoding_max_length() == 1)
-		return strlen((char *) mbstr);
+		return strlen(mbstr);
 
 	while (*mbstr)
 	{
@@ -481,17 +486,22 @@ pg_mbstrlen(const unsigned char *mbstr)
 	return (len);
 }
 
-/* returns the length (counted as a wchar) of a multibyte string
-   (not necessarily  NULL terminated) */
+/* returns the length (counted in wchars) of a multibyte string
+ * (not necessarily NULL terminated)
+ */
 int
-pg_mbstrlen_with_len(const unsigned char *mbstr, int limit)
+pg_mbstrlen_with_len(const char *mbstr, int limit)
 {
 	int			len = 0;
-	int			l;
+
+	/* optimization for single byte encoding */
+	if (pg_database_encoding_max_length() == 1)
+		return limit;
 
 	while (limit > 0 && *mbstr)
 	{
-		l = pg_mblen(mbstr);
+		int			l = pg_mblen(mbstr);
+
 		limit -= l;
 		mbstr += l;
 		len++;
@@ -506,7 +516,7 @@ pg_mbstrlen_with_len(const unsigned char *mbstr, int limit)
  * this function does not break multibyte word boundary.
  */
 int
-pg_mbcliplen(const unsigned char *mbstr, int len, int limit)
+pg_mbcliplen(const char *mbstr, int len, int limit)
 {
 	int			clen = 0;
 	int			l;
@@ -533,7 +543,7 @@ pg_mbcliplen(const unsigned char *mbstr, int len, int limit)
  * Similar to pg_mbcliplen except the limit parameter specifies the
  * character length, not the byte length.  */
 int
-pg_mbcharcliplen(const unsigned char *mbstr, int len, int limit)
+pg_mbcharcliplen(const char *mbstr, int len, int limit)
 {
 	int			clen = 0;
 	int			nch = 0;
@@ -567,7 +577,7 @@ SetDatabaseEncoding(int encoding)
 }
 
 void
-SetDefaultClientEncoding()
+SetDefaultClientEncoding(void)
 {
 	ClientEncoding = &pg_enc2name_tbl[GetDatabaseEncoding()];
 }
@@ -601,10 +611,10 @@ pg_client_encoding(PG_FUNCTION_ARGS)
 }
 
 static int
-cliplen(const unsigned char *str, int len, int limit)
+cliplen(const char *str, int len, int limit)
 {
 	int			l = 0;
-	const unsigned char *s;
+	const char *s;
 
 	for (s = str; *s; s++, l++)
 	{
