@@ -78,7 +78,7 @@
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/utils/sort/tuplesort.c,v 1.45 2004/12/31 22:02:53 pgsql Exp $
+ *	  $PostgreSQL: pgsql/src/backend/utils/sort/tuplesort.c,v 1.54 2005/10/25 13:47:08 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -94,18 +94,25 @@
 #include "utils/datum.h"
 #include "utils/logtape.h"
 #include "utils/lsyscache.h"
+#include "utils/memutils.h"
+#include "utils/pg_rusage.h"
 #include "utils/syscache.h"
 #include "utils/tuplesort.h"
 
 
+/* GUC variable */
+#ifdef TRACE_SORT
+bool		trace_sort = false;
+#endif
+
+
 /*
  * Possible states of a Tuplesort object.  These denote the states that
  * persist between calls of Tuplesort routines.
  */
 typedef enum
 {
-	TSS_INITIAL,				/* Loading tuples; still within memory
-								 * limit */
+	TSS_INITIAL,				/* Loading tuples; still within memory limit */
 	TSS_BUILDRUNS,				/* Loading tuples; writing to tape */
 	TSS_SORTEDINMEM,			/* Sort completed entirely in memory */
 	TSS_SORTEDONTAPE,			/* Sort completed, final run is on tape */
@@ -127,13 +134,13 @@ struct Tuplesortstate
 	TupSortStatus status;		/* enumerated value as shown above */
 	bool		randomAccess;	/* did caller request random access? */
 	long		availMem;		/* remaining memory available, in bytes */
-	LogicalTapeSet *tapeset;	/* logtape.c object for tapes in a temp
-								 * file */
+	long		allowedMem;		/* total memory allowed, in bytes */
+	LogicalTapeSet *tapeset;	/* logtape.c object for tapes in a temp file */
 
 	/*
-	 * These function pointers decouple the routines that must know what
-	 * kind of tuple we are sorting from the routines that don't need to
-	 * know it. They are set up by the tuplesort_begin_xxx routines.
+	 * These function pointers decouple the routines that must know what kind
+	 * of tuple we are sorting from the routines that don't need to know it.
+	 * They are set up by the tuplesort_begin_xxx routines.
 	 *
 	 * Function to compare two tuples; result is per qsort() convention, ie:
 	 *
@@ -142,83 +149,78 @@ struct Tuplesortstate
 	int			(*comparetup) (Tuplesortstate *state, const void *a, const void *b);
 
 	/*
-	 * Function to copy a supplied input tuple into palloc'd space. (NB:
-	 * we assume that a single pfree() is enough to release the tuple
-	 * later, so the representation must be "flat" in one palloc chunk.)
-	 * state->availMem must be decreased by the amount of space used.
+	 * Function to copy a supplied input tuple into palloc'd space. (NB: we
+	 * assume that a single pfree() is enough to release the tuple later, so
+	 * the representation must be "flat" in one palloc chunk.) state->availMem
+	 * must be decreased by the amount of space used.
 	 */
 	void	   *(*copytup) (Tuplesortstate *state, void *tup);
 
 	/*
-	 * Function to write a stored tuple onto tape.	The representation of
-	 * the tuple on tape need not be the same as it is in memory;
-	 * requirements on the tape representation are given below.  After
-	 * writing the tuple, pfree() it, and increase state->availMem by the
-	 * amount of memory space thereby released.
+	 * Function to write a stored tuple onto tape.	The representation of the
+	 * tuple on tape need not be the same as it is in memory; requirements on
+	 * the tape representation are given below.  After writing the tuple,
+	 * pfree() it, and increase state->availMem by the amount of memory space
+	 * thereby released.
 	 */
 	void		(*writetup) (Tuplesortstate *state, int tapenum, void *tup);
 
 	/*
-	 * Function to read a stored tuple from tape back into memory. 'len'
-	 * is the already-read length of the stored tuple.	Create and return
-	 * a palloc'd copy, and decrease state->availMem by the amount of
-	 * memory space consumed.
+	 * Function to read a stored tuple from tape back into memory. 'len' is
+	 * the already-read length of the stored tuple.  Create and return a
+	 * palloc'd copy, and decrease state->availMem by the amount of memory
+	 * space consumed.
 	 */
 	void	   *(*readtup) (Tuplesortstate *state, int tapenum, unsigned int len);
 
 	/*
-	 * This array holds pointers to tuples in sort memory.	If we are in
-	 * state INITIAL, the tuples are in no particular order; if we are in
-	 * state SORTEDINMEM, the tuples are in final sorted order; in states
-	 * BUILDRUNS and FINALMERGE, the tuples are organized in "heap" order
-	 * per Algorithm H.  (Note that memtupcount only counts the tuples
-	 * that are part of the heap --- during merge passes, memtuples[]
-	 * entries beyond TAPERANGE are never in the heap and are used to hold
-	 * pre-read tuples.)  In state SORTEDONTAPE, the array is not used.
+	 * This array holds pointers to tuples in sort memory.	If we are in state
+	 * INITIAL, the tuples are in no particular order; if we are in state
+	 * SORTEDINMEM, the tuples are in final sorted order; in states BUILDRUNS
+	 * and FINALMERGE, the tuples are organized in "heap" order per Algorithm
+	 * H.  (Note that memtupcount only counts the tuples that are part of the
+	 * heap --- during merge passes, memtuples[] entries beyond TAPERANGE are
+	 * never in the heap and are used to hold pre-read tuples.)  In state
+	 * SORTEDONTAPE, the array is not used.
 	 */
 	void	  **memtuples;		/* array of pointers to palloc'd tuples */
 	int			memtupcount;	/* number of tuples currently present */
 	int			memtupsize;		/* allocated length of memtuples array */
 
 	/*
-	 * While building initial runs, this array holds the run number for
-	 * each tuple in memtuples[].  During merge passes, we re-use it to
-	 * hold the input tape number that each tuple in the heap was read
-	 * from, or to hold the index of the next tuple pre-read from the same
-	 * tape in the case of pre-read entries.  This array is never
-	 * allocated unless we need to use tapes.  Whenever it is allocated,
-	 * it has the same length as memtuples[].
+	 * While building initial runs, this array holds the run number for each
+	 * tuple in memtuples[].  During merge passes, we re-use it to hold the
+	 * input tape number that each tuple in the heap was read from, or to hold
+	 * the index of the next tuple pre-read from the same tape in the case of
+	 * pre-read entries.  This array is never allocated unless we need to use
+	 * tapes.  Whenever it is allocated, it has the same length as
+	 * memtuples[].
 	 */
-	int		   *memtupindex;	/* index value associated with
-								 * memtuples[i] */
+	int		   *memtupindex;	/* index value associated with memtuples[i] */
 
 	/*
 	 * While building initial runs, this is the current output run number
-	 * (starting at 0).  Afterwards, it is the number of initial runs we
-	 * made.
+	 * (starting at 0).  Afterwards, it is the number of initial runs we made.
 	 */
 	int			currentRun;
 
 	/*
-	 * These variables are only used during merge passes.  mergeactive[i]
-	 * is true if we are reading an input run from (actual) tape number i
-	 * and have not yet exhausted that run.  mergenext[i] is the memtuples
-	 * index of the next pre-read tuple (next to be loaded into the heap)
-	 * for tape i, or 0 if we are out of pre-read tuples.  mergelast[i]
-	 * similarly points to the last pre-read tuple from each tape.
-	 * mergeavailmem[i] is the amount of unused space allocated for tape
-	 * i. mergefreelist and mergefirstfree keep track of unused locations
-	 * in the memtuples[] array.  memtupindex[] links together pre-read
-	 * tuples for each tape as well as recycled locations in
-	 * mergefreelist. It is OK to use 0 as a null link in these lists,
-	 * because memtuples[0] is part of the merge heap and is never a
-	 * pre-read tuple.
+	 * These variables are only used during merge passes.  mergeactive[i] is
+	 * true if we are reading an input run from (actual) tape number i and
+	 * have not yet exhausted that run.  mergenext[i] is the memtuples index
+	 * of the next pre-read tuple (next to be loaded into the heap) for tape
+	 * i, or 0 if we are out of pre-read tuples.  mergelast[i] similarly
+	 * points to the last pre-read tuple from each tape. mergeavailmem[i] is
+	 * the amount of unused space allocated for tape i. mergefreelist and
+	 * mergefirstfree keep track of unused locations in the memtuples[] array.
+	 * memtupindex[] links together pre-read tuples for each tape as well as
+	 * recycled locations in mergefreelist. It is OK to use 0 as a null link
+	 * in these lists, because memtuples[0] is part of the merge heap and is
+	 * never a pre-read tuple.
 	 */
 	bool		mergeactive[MAXTAPES];	/* Active input run source? */
-	int			mergenext[MAXTAPES];	/* first preread tuple for each
-										 * source */
-	int			mergelast[MAXTAPES];	/* last preread tuple for each
-										 * source */
+	int			mergenext[MAXTAPES];	/* first preread tuple for each source */
+	int			mergelast[MAXTAPES];	/* last preread tuple for each source */
 	long		mergeavailmem[MAXTAPES];		/* availMem for prereading
 												 * tapes */
 	long		spacePerTape;	/* actual per-tape target usage */
@@ -232,17 +234,15 @@ struct Tuplesortstate
 	 */
 	int			Level;			/* Knuth's l */
 	int			destTape;		/* current output tape (Knuth's j, less 1) */
-	int			tp_fib[MAXTAPES];		/* Target Fibonacci run counts
-										 * (A[]) */
+	int			tp_fib[MAXTAPES];		/* Target Fibonacci run counts (A[]) */
 	int			tp_runs[MAXTAPES];		/* # of real runs on each tape */
-	int			tp_dummy[MAXTAPES];		/* # of dummy runs for each tape
-										 * (D[]) */
+	int			tp_dummy[MAXTAPES];		/* # of dummy runs for each tape (D[]) */
 	int			tp_tapenum[MAXTAPES];	/* Actual tape numbers (TAPE[]) */
 
 	/*
-	 * These variables are used after completion of sorting to keep track
-	 * of the next tuple to return.  (In the tape case, the tape's current
-	 * read position is also critical state.)
+	 * These variables are used after completion of sorting to keep track of
+	 * the next tuple to return.  (In the tape case, the tape's current read
+	 * position is also critical state.)
 	 */
 	int			result_tape;	/* actual tape number of finished output */
 	int			current;		/* array index (only used if SORTEDINMEM) */
@@ -250,8 +250,7 @@ struct Tuplesortstate
 
 	/* markpos_xxx holds marked position for mark and restore */
 	long		markpos_block;	/* tape block# (only used if SORTEDONTAPE) */
-	int			markpos_offset; /* saved "current", or offset in tape
-								 * block */
+	int			markpos_offset; /* saved "current", or offset in tape block */
 	bool		markpos_eof;	/* saved "eof_reached" */
 
 	/*
@@ -264,8 +263,8 @@ struct Tuplesortstate
 	SortFunctionKind *sortFnKinds;
 
 	/*
-	 * These variables are specific to the IndexTuple case; they are set
-	 * by tuplesort_begin_index and used only by the IndexTuple routines.
+	 * These variables are specific to the IndexTuple case; they are set by
+	 * tuplesort_begin_index and used only by the IndexTuple routines.
 	 */
 	Relation	indexRel;
 	ScanKey		indexScanKey;
@@ -282,6 +281,13 @@ struct Tuplesortstate
 	/* we need typelen and byval in order to know how to copy the Datums. */
 	int			datumTypeLen;
 	bool		datumTypeByVal;
+
+	/*
+	 * Resource snapshot for time of sort start.
+	 */
+#ifdef TRACE_SORT
+	PGRUsage	ru_start;
+#endif
 };
 
 #define COMPARETUP(state,a,b)	((*(state)->comparetup) (state, a, b))
@@ -421,9 +427,15 @@ tuplesort_begin_common(int workMem, bool randomAccess)
 
 	state = (Tuplesortstate *) palloc0(sizeof(Tuplesortstate));
 
+#ifdef TRACE_SORT
+	if (trace_sort)
+		pg_rusage_init(&state->ru_start);
+#endif
+
 	state->status = TSS_INITIAL;
 	state->randomAccess = randomAccess;
-	state->availMem = workMem * 1024L;
+	state->allowedMem = workMem * 1024L;
+	state->availMem = state->allowedMem;
 	state->tapeset = NULL;
 
 	state->memtupcount = 0;
@@ -438,8 +450,7 @@ tuplesort_begin_common(int workMem, bool randomAccess)
 
 	/* Algorithm D variables will be initialized by inittapes, if needed */
 
-	state->result_tape = -1;	/* flag that result tape has not been
-								 * formed */
+	state->result_tape = -1;	/* flag that result tape has not been formed */
 
 	return state;
 }
@@ -455,6 +466,13 @@ tuplesort_begin_heap(TupleDesc tupDesc,
 
 	AssertArg(nkeys > 0);
 
+#ifdef TRACE_SORT
+	if (trace_sort)
+		elog(LOG,
+			 "begin tuple sort: nkeys = %d, workMem = %d, randomAccess = %c",
+			 nkeys, workMem, randomAccess ? 't' : 'f');
+#endif
+
 	state->comparetup = comparetup_heap;
 	state->copytup = copytup_heap;
 	state->writetup = writetup_heap;
@@ -478,8 +496,8 @@ tuplesort_begin_heap(TupleDesc tupDesc,
 						   &state->sortFnKinds[i]);
 
 		/*
-		 * We needn't fill in sk_strategy or sk_subtype since these
-		 * scankeys will never be passed to an index.
+		 * We needn't fill in sk_strategy or sk_subtype since these scankeys
+		 * will never be passed to an index.
 		 */
 		ScanKeyInit(&state->scanKeys[i],
 					attNums[i],
@@ -498,6 +516,14 @@ tuplesort_begin_index(Relation indexRel,
 {
 	Tuplesortstate *state = tuplesort_begin_common(workMem, randomAccess);
 
+#ifdef TRACE_SORT
+	if (trace_sort)
+		elog(LOG,
+			 "begin index sort: unique = %c, workMem = %d, randomAccess = %c",
+			 enforceUnique ? 't' : 'f',
+			 workMem, randomAccess ? 't' : 'f');
+#endif
+
 	state->comparetup = comparetup_index;
 	state->copytup = copytup_index;
 	state->writetup = writetup_index;
@@ -521,6 +547,13 @@ tuplesort_begin_datum(Oid datumType,
 	int16		typlen;
 	bool		typbyval;
 
+#ifdef TRACE_SORT
+	if (trace_sort)
+		elog(LOG,
+			 "begin datum sort: workMem = %d, randomAccess = %c",
+			 workMem, randomAccess ? 't' : 'f');
+#endif
+
 	state->comparetup = comparetup_datum;
 	state->copytup = copytup_datum;
 	state->writetup = writetup_datum;
@@ -551,9 +584,24 @@ void
 tuplesort_end(Tuplesortstate *state)
 {
 	int			i;
+#ifdef TRACE_SORT
+	long		spaceUsed;
+#endif
 
 	if (state->tapeset)
+	{
+#ifdef TRACE_SORT
+		spaceUsed = LogicalTapeSetBlocks(state->tapeset);
+#endif
 		LogicalTapeSetClose(state->tapeset);
+	}
+	else
+	{
+#ifdef TRACE_SORT
+		spaceUsed = (state->allowedMem - state->availMem + 1023) / 1024;
+#endif
+	}
+
 	if (state->memtuples)
 	{
 		for (i = 0; i < state->memtupcount; i++)
@@ -564,14 +612,25 @@ tuplesort_end(Tuplesortstate *state)
 		pfree(state->memtupindex);
 
 	/*
-	 * this stuff might better belong in a variant-specific shutdown
-	 * routine
+	 * this stuff might better belong in a variant-specific shutdown routine
 	 */
 	if (state->scanKeys)
 		pfree(state->scanKeys);
 	if (state->sortFnKinds)
 		pfree(state->sortFnKinds);
 
+#ifdef TRACE_SORT
+	if (trace_sort)
+	{
+		if (state->tapeset)
+			elog(LOG, "external sort ended, %ld disk blocks used: %s",
+				 spaceUsed, pg_rusage_show(&state->ru_start));
+		else
+			elog(LOG, "internal sort ended, %ld KB used: %s",
+				 spaceUsed, pg_rusage_show(&state->ru_start));
+	}
+#endif
+
 	pfree(state);
 }
 
@@ -676,16 +735,16 @@ puttuple_common(Tuplesortstate *state, void *tuple)
 
 			/*
 			 * Insert the copied tuple into the heap, with run number
-			 * currentRun if it can go into the current run, else run
-			 * number currentRun+1.  The tuple can go into the current run
-			 * if it is >= the first not-yet-output tuple.	(Actually, it
-			 * could go into the current run if it is >= the most recently
-			 * output tuple ... but that would require keeping around the
-			 * tuple we last output, and it's simplest to let writetup
-			 * free each tuple as soon as it's written.)
+			 * currentRun if it can go into the current run, else run number
+			 * currentRun+1.  The tuple can go into the current run if it is
+			 * >= the first not-yet-output tuple.  (Actually, it could go into
+			 * the current run if it is >= the most recently output tuple ...
+			 * but that would require keeping around the tuple we last output,
+			 * and it's simplest to let writetup free each tuple as soon as
+			 * it's written.)
 			 *
-			 * Note there will always be at least one tuple in the heap at
-			 * this point; see dumptuples.
+			 * Note there will always be at least one tuple in the heap at this
+			 * point; see dumptuples.
 			 */
 			Assert(state->memtupcount > 0);
 			if (COMPARETUP(state, tuple, state->memtuples[0]) >= 0)
@@ -694,8 +753,7 @@ puttuple_common(Tuplesortstate *state, void *tuple)
 				tuplesort_heap_insert(state, tuple, state->currentRun + 1, true);
 
 			/*
-			 * If we are over the memory limit, dump tuples till we're
-			 * under.
+			 * If we are over the memory limit, dump tuples till we're under.
 			 */
 			dumptuples(state, false);
 			break;
@@ -711,13 +769,19 @@ puttuple_common(Tuplesortstate *state, void *tuple)
 void
 tuplesort_performsort(Tuplesortstate *state)
 {
+#ifdef TRACE_SORT
+	if (trace_sort)
+		elog(LOG, "performsort starting: %s",
+			 pg_rusage_show(&state->ru_start));
+#endif
+
 	switch (state->status)
 	{
 		case TSS_INITIAL:
 
 			/*
-			 * We were able to accumulate all the tuples within the
-			 * allowed amount of memory.  Just qsort 'em and we're done.
+			 * We were able to accumulate all the tuples within the allowed
+			 * amount of memory.  Just qsort 'em and we're done.
 			 */
 			if (state->memtupcount > 1)
 			{
@@ -734,10 +798,10 @@ tuplesort_performsort(Tuplesortstate *state)
 		case TSS_BUILDRUNS:
 
 			/*
-			 * Finish tape-based sort.	First, flush all tuples remaining
-			 * in memory out to tape; then merge until we have a single
-			 * remaining run (or, if !randomAccess, one run per tape).
-			 * Note that mergeruns sets the correct state->status.
+			 * Finish tape-based sort.	First, flush all tuples remaining in
+			 * memory out to tape; then merge until we have a single remaining
+			 * run (or, if !randomAccess, one run per tape). Note that
+			 * mergeruns sets the correct state->status.
 			 */
 			dumptuples(state, true);
 			mergeruns(state);
@@ -750,6 +814,13 @@ tuplesort_performsort(Tuplesortstate *state)
 			elog(ERROR, "invalid tuplesort state");
 			break;
 	}
+
+#ifdef TRACE_SORT
+	if (trace_sort)
+		elog(LOG, "performsort done%s: %s",
+			 (state->status == TSS_FINALMERGE) ? " (except final merge)" : "",
+			 pg_rusage_show(&state->ru_start));
+#endif
 }
 
 /*
@@ -819,16 +890,15 @@ tuplesort_gettuple(Tuplesortstate *state, bool forward,
 			/*
 			 * Backward.
 			 *
-			 * if all tuples are fetched already then we return last tuple,
-			 * else - tuple before last returned.
+			 * if all tuples are fetched already then we return last tuple, else
+			 * - tuple before last returned.
 			 */
 			if (state->eof_reached)
 			{
 				/*
-				 * Seek position is pointing just past the zero tuplen at
-				 * the end of file; back up to fetch last tuple's ending
-				 * length word.  If seek fails we must have a completely
-				 * empty file.
+				 * Seek position is pointing just past the zero tuplen at the
+				 * end of file; back up to fetch last tuple's ending length
+				 * word.  If seek fails we must have a completely empty file.
 				 */
 				if (!LogicalTapeBackspace(state->tapeset,
 										  state->result_tape,
@@ -839,9 +909,8 @@ tuplesort_gettuple(Tuplesortstate *state, bool forward,
 			else
 			{
 				/*
-				 * Back up and fetch previously-returned tuple's ending
-				 * length word.  If seek fails, assume we are at start of
-				 * file.
+				 * Back up and fetch previously-returned tuple's ending length
+				 * word.  If seek fails, assume we are at start of file.
 				 */
 				if (!LogicalTapeBackspace(state->tapeset,
 										  state->result_tape,
@@ -854,17 +923,17 @@ tuplesort_gettuple(Tuplesortstate *state, bool forward,
 				 */
 				if (!LogicalTapeBackspace(state->tapeset,
 										  state->result_tape,
-									  tuplen + 2 * sizeof(unsigned int)))
+										  tuplen + 2 * sizeof(unsigned int)))
 				{
 					/*
-					 * If that fails, presumably the prev tuple is the
-					 * first in the file.  Back up so that it becomes next
-					 * to read in forward direction (not obviously right,
-					 * but that is what in-memory case does).
+					 * If that fails, presumably the prev tuple is the first
+					 * in the file.  Back up so that it becomes next to read
+					 * in forward direction (not obviously right, but that is
+					 * what in-memory case does).
 					 */
 					if (!LogicalTapeBackspace(state->tapeset,
 											  state->result_tape,
-										  tuplen + sizeof(unsigned int)))
+											  tuplen + sizeof(unsigned int)))
 						elog(ERROR, "bogus tuple length in backward scan");
 					return NULL;
 				}
@@ -873,9 +942,9 @@ tuplesort_gettuple(Tuplesortstate *state, bool forward,
 			tuplen = getlen(state, state->result_tape, false);
 
 			/*
-			 * Now we have the length of the prior tuple, back up and read
-			 * it. Note: READTUP expects we are positioned after the
-			 * initial length word of the tuple, so back up to that point.
+			 * Now we have the length of the prior tuple, back up and read it.
+			 * Note: READTUP expects we are positioned after the initial
+			 * length word of the tuple, so back up to that point.
 			 */
 			if (!LogicalTapeBackspace(state->tapeset,
 									  state->result_tape,
@@ -907,14 +976,12 @@ tuplesort_gettuple(Tuplesortstate *state, bool forward,
 				if ((tupIndex = state->mergenext[srcTape]) == 0)
 				{
 					/*
-					 * out of preloaded data on this tape, try to read
-					 * more
+					 * out of preloaded data on this tape, try to read more
 					 */
 					mergepreread(state);
 
 					/*
-					 * if still no data, we've reached end of run on this
-					 * tape
+					 * if still no data, we've reached end of run on this tape
 					 */
 					if ((tupIndex = state->mergenext[srcTape]) == 0)
 						return tup;
@@ -985,6 +1052,12 @@ inittapes(Tuplesortstate *state)
 	int			ntuples,
 				j;
 
+#ifdef TRACE_SORT
+	if (trace_sort)
+		elog(LOG, "switching to external sort: %s",
+			 pg_rusage_show(&state->ru_start));
+#endif
+
 	state->tapeset = LogicalTapeSetCreate(MAXTAPES);
 
 	/*
@@ -995,12 +1068,12 @@ inittapes(Tuplesortstate *state)
 	USEMEM(state, GetMemoryChunkSpace(state->memtupindex));
 
 	/*
-	 * Convert the unsorted contents of memtuples[] into a heap. Each
-	 * tuple is marked as belonging to run number zero.
+	 * Convert the unsorted contents of memtuples[] into a heap. Each tuple is
+	 * marked as belonging to run number zero.
 	 *
 	 * NOTE: we pass false for checkIndex since there's no point in comparing
-	 * indexes in this step, even though we do intend the indexes to be
-	 * part of the sort key...
+	 * indexes in this step, even though we do intend the indexes to be part
+	 * of the sort key...
 	 */
 	ntuples = state->memtupcount;
 	state->memtupcount = 0;		/* make the heap empty */
@@ -1083,8 +1156,8 @@ mergeruns(Tuplesortstate *state)
 
 	/*
 	 * If we produced only one initial run (quite likely if the total data
-	 * volume is between 1X and 2X workMem), we can just use that tape as
-	 * the finished output, rather than doing a useless merge.
+	 * volume is between 1X and 2X workMem), we can just use that tape as the
+	 * finished output, rather than doing a useless merge.
 	 */
 	if (state->currentRun == 1)
 	{
@@ -1116,8 +1189,8 @@ mergeruns(Tuplesortstate *state)
 			}
 
 			/*
-			 * If we don't have to produce a materialized sorted tape,
-			 * quit as soon as we're down to one real/dummy run per tape.
+			 * If we don't have to produce a materialized sorted tape, quit as
+			 * soon as we're down to one real/dummy run per tape.
 			 */
 			if (!state->randomAccess && allOneRun)
 			{
@@ -1148,8 +1221,7 @@ mergeruns(Tuplesortstate *state)
 		state->tp_runs[TAPERANGE - 1] = 0;
 
 		/*
-		 * reassign tape units per step D6; note we no longer care about
-		 * A[]
+		 * reassign tape units per step D6; note we no longer care about A[]
 		 */
 		svTape = state->tp_tapenum[TAPERANGE];
 		svDummy = state->tp_dummy[TAPERANGE];
@@ -1166,12 +1238,12 @@ mergeruns(Tuplesortstate *state)
 	}
 
 	/*
-	 * Done.  Knuth says that the result is on TAPE[1], but since we
-	 * exited the loop without performing the last iteration of step D6,
-	 * we have not rearranged the tape unit assignment, and therefore the
-	 * result is on TAPE[T].  We need to do it this way so that we can
-	 * freeze the final output tape while rewinding it.  The last
-	 * iteration of step D6 would be a waste of cycles anyway...
+	 * Done.  Knuth says that the result is on TAPE[1], but since we exited
+	 * the loop without performing the last iteration of step D6, we have not
+	 * rearranged the tape unit assignment, and therefore the result is on
+	 * TAPE[T].  We need to do it this way so that we can freeze the final
+	 * output tape while rewinding it.	The last iteration of step D6 would be
+	 * a waste of cycles anyway...
 	 */
 	state->result_tape = state->tp_tapenum[TAPERANGE];
 	LogicalTapeFreeze(state->tapeset, state->result_tape);
@@ -1195,16 +1267,15 @@ mergeonerun(Tuplesortstate *state)
 				spaceFreed;
 
 	/*
-	 * Start the merge by loading one tuple from each active source tape
-	 * into the heap.  We can also decrease the input run/dummy run
-	 * counts.
+	 * Start the merge by loading one tuple from each active source tape into
+	 * the heap.  We can also decrease the input run/dummy run counts.
 	 */
 	beginmerge(state);
 
 	/*
-	 * Execute merge by repeatedly extracting lowest tuple in heap,
-	 * writing it out, and replacing it with next tuple from same tape (if
-	 * there is another one).
+	 * Execute merge by repeatedly extracting lowest tuple in heap, writing it
+	 * out, and replacing it with next tuple from same tape (if there is
+	 * another one).
 	 */
 	while (state->memtupcount > 0)
 	{
@@ -1237,11 +1308,17 @@ mergeonerun(Tuplesortstate *state)
 	}
 
 	/*
-	 * When the heap empties, we're done.  Write an end-of-run marker on
-	 * the output tape, and increment its count of real runs.
+	 * When the heap empties, we're done.  Write an end-of-run marker on the
+	 * output tape, and increment its count of real runs.
 	 */
 	markrunend(state, destTape);
 	state->tp_runs[TAPERANGE]++;
+
+#ifdef TRACE_SORT
+	if (trace_sort)
+		elog(LOG, "finished merge step: %s",
+			 pg_rusage_show(&state->ru_start));
+#endif
 }
 
 /*
@@ -1268,8 +1345,7 @@ beginmerge(Tuplesortstate *state)
 	memset(state->mergelast, 0, sizeof(state->mergelast));
 	memset(state->mergeavailmem, 0, sizeof(state->mergeavailmem));
 	state->mergefreelist = 0;	/* nothing in the freelist */
-	state->mergefirstfree = MAXTAPES;	/* first slot available for
-										 * preread */
+	state->mergefirstfree = MAXTAPES;	/* first slot available for preread */
 
 	/* Adjust run counts and mark the active tapes */
 	activeTapes = 0;
@@ -1288,8 +1364,8 @@ beginmerge(Tuplesortstate *state)
 	}
 
 	/*
-	 * Initialize space allocation to let each active input tape have an
-	 * equal share of preread space.
+	 * Initialize space allocation to let each active input tape have an equal
+	 * share of preread space.
 	 */
 	Assert(activeTapes > 0);
 	state->spacePerTape = state->availMem / activeTapes;
@@ -1300,8 +1376,8 @@ beginmerge(Tuplesortstate *state)
 	}
 
 	/*
-	 * Preread as many tuples as possible (and at least one) from each
-	 * active tape
+	 * Preread as many tuples as possible (and at least one) from each active
+	 * tape
 	 */
 	mergepreread(state);
 
@@ -1331,7 +1407,7 @@ beginmerge(Tuplesortstate *state)
  * as explained in the header comments of this file.  Load tuples from each
  * active source tape until the tape's run is exhausted or it has used up
  * its fair share of available memory.	In any case, we guarantee that there
- * is at one preread tuple available from each unexhausted input tape.
+ * is at least one preread tuple available from each unexhausted input tape.
  */
 static void
 mergepreread(Tuplesortstate *state)
@@ -1359,8 +1435,8 @@ mergepreread(Tuplesortstate *state)
 			continue;
 
 		/*
-		 * Read tuples from this tape until it has used up its free
-		 * memory, but ensure that we have at least one.
+		 * Read tuples from this tape until it has used up its free memory,
+		 * but ensure that we have at least one.
 		 */
 		priorAvail = state->availMem;
 		state->availMem = state->mergeavailmem[srcTape];
@@ -1435,8 +1511,8 @@ dumptuples(Tuplesortstate *state, bool alltuples)
 		   (LACKMEM(state) && state->memtupcount > 1))
 	{
 		/*
-		 * Dump the heap's frontmost entry, and sift up to remove it from
-		 * the heap.
+		 * Dump the heap's frontmost entry, and sift up to remove it from the
+		 * heap.
 		 */
 		Assert(state->memtupcount > 0);
 		WRITETUP(state, state->tp_tapenum[state->destTape],
@@ -1455,6 +1531,14 @@ dumptuples(Tuplesortstate *state, bool alltuples)
 			state->tp_runs[state->destTape]++;
 			state->tp_dummy[state->destTape]--; /* per Alg D step D2 */
 
+#ifdef TRACE_SORT
+			if (trace_sort)
+				elog(LOG, "finished writing%s run %d: %s",
+					 (state->memtupcount == 0) ? " final" : "",
+					 state->currentRun,
+					 pg_rusage_show(&state->ru_start));
+#endif
+
 			/*
 			 * Done if heap is empty, else prepare for new run.
 			 */
@@ -1563,7 +1647,7 @@ tuplesort_restorepos(Tuplesortstate *state)
  */
 
 #define HEAPCOMPARE(tup1,index1,tup2,index2) \
-	(checkIndex && (index1 != index2) ? index1 - index2 : \
+	(checkIndex && (index1 != index2) ? (index1) - (index2) : \
 	 COMPARETUP(state, tup1, tup2))
 
 /*
@@ -1599,8 +1683,8 @@ tuplesort_heap_insert(Tuplesortstate *state, void *tuple,
 	memtupindex = state->memtupindex;
 
 	/*
-	 * Sift-up the new entry, per Knuth 5.2.3 exercise 16. Note that Knuth
-	 * is using 1-based array indexes, not 0-based.
+	 * Sift-up the new entry, per Knuth 5.2.3 exercise 16. Note that Knuth is
+	 * using 1-based array indexes, not 0-based.
 	 */
 	j = state->memtupcount++;
 	while (j > 0)
@@ -1724,12 +1808,12 @@ SelectSortFunction(Oid sortOperator,
 	Oid			opclass = InvalidOid;
 
 	/*
-	 * Search pg_amop to see if the target operator is registered as the
-	 * "<" or ">" operator of any btree opclass.  It's possible that it
-	 * might be registered both ways (eg, if someone were to build a
-	 * "reverse sort" opclass for some reason); prefer the "<" case if so.
-	 * If the operator is registered the same way in multiple opclasses,
-	 * assume we can use the associated comparator function from any one.
+	 * Search pg_amop to see if the target operator is registered as the "<"
+	 * or ">" operator of any btree opclass.  It's possible that it might be
+	 * registered both ways (eg, if someone were to build a "reverse sort"
+	 * opclass for some reason); prefer the "<" case if so. If the operator is
+	 * registered the same way in multiple opclasses, assume we can use the
+	 * associated comparator function from any one.
 	 */
 	catlist = SearchSysCacheList(AMOPOPID, 1,
 								 ObjectIdGetDatum(sortOperator),
@@ -1773,11 +1857,11 @@ SelectSortFunction(Oid sortOperator,
 	}
 
 	/*
-	 * Can't find a comparator, so use the operator as-is.  Decide whether
-	 * it is forward or reverse sort by looking at its name (grotty, but
-	 * this only matters for deciding which end NULLs should get sorted
-	 * to).  XXX possibly better idea: see whether its selectivity
-	 * function is scalargtcmp?
+	 * Can't find a comparator, so use the operator as-is.  Decide whether it
+	 * is forward or reverse sort by looking at its name (grotty, but this
+	 * only matters for deciding which end NULLs should get sorted to).  XXX
+	 * possibly better idea: see whether its selectivity function is
+	 * scalargtcmp?
 	 */
 	tuple = SearchSysCache(OPEROID,
 						   ObjectIdGetDatum(sortOperator),
@@ -1804,13 +1888,8 @@ myFunctionCall2(FmgrInfo *flinfo, Datum arg1, Datum arg2)
 	FunctionCallInfoData fcinfo;
 	Datum		result;
 
-	/* MemSet(&fcinfo, 0, sizeof(fcinfo)); */
-	fcinfo.context = NULL;
-	fcinfo.resultinfo = NULL;
-	fcinfo.isnull = false;
+	InitFunctionCallInfoData(fcinfo, flinfo, 2, NULL, NULL);
 
-	fcinfo.flinfo = flinfo;
-	fcinfo.nargs = 2;
 	fcinfo.arg[0] = arg1;
 	fcinfo.arg[1] = arg2;
 	fcinfo.argnull[0] = false;
@@ -2066,15 +2145,15 @@ comparetup_index(Tuplesortstate *state, const void *a, const void *b)
 	 * If btree has asked us to enforce uniqueness, complain if two equal
 	 * tuples are detected (unless there was at least one NULL field).
 	 *
-	 * It is sufficient to make the test here, because if two tuples are
-	 * equal they *must* get compared at some stage of the sort ---
-	 * otherwise the sort algorithm wouldn't have checked whether one must
-	 * appear before the other.
+	 * It is sufficient to make the test here, because if two tuples are equal
+	 * they *must* get compared at some stage of the sort --- otherwise the
+	 * sort algorithm wouldn't have checked whether one must appear before the
+	 * other.
 	 *
-	 * Some rather brain-dead implementations of qsort will sometimes call
-	 * the comparison routine to compare a value to itself.  (At this
-	 * writing only QNX 4 is known to do such silly things.)  Don't raise
-	 * a bogus error in that case.
+	 * Some rather brain-dead implementations of qsort will sometimes call the
+	 * comparison routine to compare a value to itself.  (At this writing only
+	 * QNX 4 is known to do such silly things.)  Don't raise a bogus error in
+	 * that case.
 	 */
 	if (state->enforceUnique && !equal_hasnull && tuple1 != tuple2)
 		ereport(ERROR,
@@ -2083,10 +2162,10 @@ comparetup_index(Tuplesortstate *state, const void *a, const void *b)
 				 errdetail("Table contains duplicated values.")));
 
 	/*
-	 * If key values are equal, we sort on ItemPointer.  This does not
-	 * affect validity of the finished index, but it offers cheap
-	 * insurance against performance problems with bad qsort
-	 * implementations that have trouble with large numbers of equal keys.
+	 * If key values are equal, we sort on ItemPointer.  This does not affect
+	 * validity of the finished index, but it offers cheap insurance against
+	 * performance problems with bad qsort implementations that have trouble
+	 * with large numbers of equal keys.
 	 */
 	{
 		BlockNumber blk1 = ItemPointerGetBlockNumber(&tuple1->t_tid);
