@@ -3,10 +3,10 @@
  * crypt.h
  *	  Interface to libpq/crypt.c
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
- * $Id: crypt.h,v 1.22.2.1 2002/12/05 18:40:08 momjian Exp $
+ * $PostgreSQL: pgsql/src/include/libpq/crypt.h,v 1.32 2005/10/17 16:24:20 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -15,20 +15,19 @@
 
 #include "libpq/libpq-be.h"
 
-/* Also defined in interfaces/odbc/md5.h */
 #define MD5_PASSWD_LEN	35
 
-#define isMD5(passwd)	(strncmp((passwd),"md5",3) == 0 && \
+#define isMD5(passwd)	(strncmp(passwd, "md5", 3) == 0 && \
 						 strlen(passwd) == MD5_PASSWD_LEN)
 
 
+/* in crypt.c */
 extern int md5_crypt_verify(const Port *port, const char *user,
-				char *pgpass);
-extern bool md5_hash(const void *buff, size_t len, char *hexsum);
-extern bool CheckMD5Pwd(char *passwd, char *storedpwd, char *seed);
+				 char *client_pass);
 
-/* Also defined in interfaces/odbc/md5.h */
-extern bool EncryptMD5(const char *passwd, const char *salt,
+/* in md5.c --- these are also present in frontend libpq */
+extern bool pg_md5_hash(const void *buff, size_t len, char *hexsum);
+extern bool pg_md5_encrypt(const char *passwd, const char *salt,
 		   size_t salt_len, char *buf);
 
 #endif
